from __future__ import annotations

# 책임 경계별 타입 계층을 재노출한다.
from .lsp_types_base import *  # noqa: F401,F403
from .lsp_types_protocol import *  # noqa: F401,F403
from .lsp_types_capabilities import *  # noqa: F401,F403



"""MCP 도구 노출 정책을 정의한다."""

from __future__ import annotations

from copy import deepcopy


HIDDEN_TOOL_NAMES: frozenset[str] = frozenset(
    {
        "save_snippet",
        "get_snippet",
        "archive_context",
        "get_context",
        "pipeline_policy_get",
        "pipeline_policy_set",
        "pipeline_alert_status",
        "pipeline_dead_list",
        "pipeline_dead_requeue",
        "pipeline_dead_purge",
        "pipeline_auto_status",
        "pipeline_auto_set",
        "pipeline_auto_tick",
        "pipeline_benchmark_run",
        "pipeline_benchmark_report",
        "pipeline_quality_run",
        "pipeline_quality_report",
        "pipeline_lsp_matrix_run",
        "pipeline_lsp_matrix_report",
    }
)


def is_hidden_tool_name(tool_name: object) -> bool:
    """지정 도구명이 숨김 정책 대상인지 반환한다."""
    return isinstance(tool_name, str) and tool_name in HIDDEN_TOOL_NAMES


def filter_tools_list_response_payload(payload: dict[str, object]) -> dict[str, object]:
    """tools/list JSON-RPC 응답에서 숨김 도구를 제거한다."""
    result_obj = payload.get("result")
    if not isinstance(result_obj, dict):
        return payload
    tools_obj = result_obj.get("tools")
    if not isinstance(tools_obj, list):
        return payload
    filtered_tools = [tool for tool in tools_obj if not _is_hidden_tool_entry(tool)]
    if len(filtered_tools) == len(tools_obj):
        return payload
    copied_payload = deepcopy(payload)
    copied_result = copied_payload.get("result")
    if isinstance(copied_result, dict):
        copied_result["tools"] = filtered_tools
    return copied_payload


def _is_hidden_tool_entry(entry: object) -> bool:
    """tools/list 항목이 숨김 대상인지 판정한다."""
    if not isinstance(entry, dict):
        return False
    return is_hidden_tool_name(entry.get("name"))



"""
This file contains various utility functions like I/O operations, handling paths, etc.
"""

import gzip
import logging
import os
import platform
import shutil
import subprocess
import uuid
import zipfile
from enum import Enum
from pathlib import Path, PurePath

import charset_normalizer
import requests

from solidlsp.ls_exceptions import SolidLSPException
from solidlsp.ls_types import UnifiedSymbolInformation

log = logging.getLogger(__name__)


class InvalidTextLocationError(Exception):
    pass


class TextUtils:
    """
    Utilities for text operations.
    """

    @staticmethod
    def get_line_col_from_index(text: str, index: int) -> tuple[int, int]:
        """
        Returns the zero-indexed line and column number of the given index in the given text
        """
        l = 0
        c = 0
        idx = 0
        while idx < index:
            if text[idx] == "\n":
                l += 1
                c = 0
            else:
                c += 1
            idx += 1

        return l, c

    @staticmethod
    def get_index_from_line_col(text: str, line: int, col: int) -> int:
        """
        Returns the index of the given zero-indexed line and column number in the given text
        """
        idx = 0
        while line > 0:
            if idx >= len(text):
                raise InvalidTextLocationError
            if text[idx] == "\n":
                line -= 1
            idx += 1
        idx += col
        return idx

    @staticmethod
    def _get_updated_position_from_line_and_column_and_edit(l: int, c: int, text_to_be_inserted: str) -> tuple[int, int]:
        """
        Utility function to get the position of the cursor after inserting text at a given line and column.
        """
        num_newlines_in_gen_text = text_to_be_inserted.count("\n")
        if num_newlines_in_gen_text > 0:
            l += num_newlines_in_gen_text
            c = len(text_to_be_inserted.split("\n")[-1])
        else:
            c += len(text_to_be_inserted)
        return (l, c)

    @staticmethod
    def delete_text_between_positions(text: str, start_line: int, start_col: int, end_line: int, end_col: int) -> tuple[str, str]:
        """
        Deletes the text between the given start and end positions.
        Returns the modified text and the deleted text.
        """
        del_start_idx = TextUtils.get_index_from_line_col(text, start_line, start_col)
        del_end_idx = TextUtils.get_index_from_line_col(text, end_line, end_col)

        deleted_text = text[del_start_idx:del_end_idx]
        new_text = text[:del_start_idx] + text[del_end_idx:]
        return new_text, deleted_text

    @staticmethod
    def insert_text_at_position(text: str, line: int, col: int, text_to_be_inserted: str) -> tuple[str, int, int]:
        """
        Inserts the given text at the given line and column.
        Returns the modified text and the new line and column.
        """
        try:
            change_index = TextUtils.get_index_from_line_col(text, line, col)
        except InvalidTextLocationError:
            num_lines_in_text = text.count("\n") + 1
            max_line = num_lines_in_text - 1
            if line == max_line + 1 and col == 0:  # trying to insert at new line after full text
                # insert at end, adding missing newline
                change_index = len(text)
                text_to_be_inserted = "\n" + text_to_be_inserted
            else:
                raise
        new_text = text[:change_index] + text_to_be_inserted + text[change_index:]
        new_l, new_c = TextUtils._get_updated_position_from_line_and_column_and_edit(line, col, text_to_be_inserted)
        return new_text, new_l, new_c


class PathUtils:
    """
    Utilities for platform-agnostic path operations.
    """

    @staticmethod
    def uri_to_path(uri: str) -> str:
        """
        Converts a URI to a file path. Works on both Linux and Windows.

        This method was obtained from https://stackoverflow.com/a/61922504
        """
        try:
            from urllib.parse import unquote, urlparse
            from urllib.request import url2pathname
        except ImportError:
            # backwards compatibility (Python 2)
            from urllib.parse import unquote as unquote_py2
            from urllib.request import url2pathname as url2pathname_py2

            from urlparse import urlparse as urlparse_py2

            unquote = unquote_py2
            url2pathname = url2pathname_py2
            urlparse = urlparse_py2
        parsed = urlparse(uri)
        host = f"{os.path.sep}{os.path.sep}{parsed.netloc}{os.path.sep}"
        path = os.path.normpath(os.path.join(host, url2pathname(unquote(parsed.path))))
        return path

    @staticmethod
    def path_to_uri(path: str) -> str:
        """
        Converts a file path to a file URI (file:///...).
        """
        return str(Path(path).absolute().as_uri())

    @staticmethod
    def is_glob_pattern(pattern: str) -> bool:
        """Check if a pattern contains glob-specific characters."""
        return any(c in pattern for c in "*?[]!")

    @staticmethod
    def get_relative_path(path: str, base_path: str) -> str | None:
        """
        Gets relative path if it's possible (paths should be on the same drive),
        returns `None` otherwise.
        """
        if PurePath(path).drive == PurePath(base_path).drive:
            rel_path = str(PurePath(os.path.relpath(path, base_path)))
            return rel_path
        return None


class FileUtils:
    """
    Utility functions for file operations.
    """

    @staticmethod
    def read_file(file_path: str, encoding: str) -> str:
        """
        Reads the file at the given path using the given encoding and returns the contents as a string.
        If decoding fails, tries to detect the encoding using charset_normalizer.

        Raises FileNotFoundError if the file does not exist.
        """
        if not os.path.exists(file_path):
            log.error(f"Failed to read '{file_path}': File does not exist.")
            raise FileNotFoundError(f"File read '{file_path}' failed: File does not exist.")
        try:
            try:
                with open(file_path, encoding=encoding) as inp_file:
                    return inp_file.read()
            except UnicodeDecodeError as ude:
                results = charset_normalizer.from_path(file_path)
                match = results.best()
                if match:
                    log.warning(
                        f"Could not decode {file_path} with encoding='{encoding}'; using best match '{match.encoding}' instead",
                    )
                    return match.raw.decode(match.encoding)
                raise ude
        except (OSError, UnicodeError) as exc:
            log.error(f"Failed to read '{file_path}' with encoding '{encoding}': {exc}")
            raise exc

    @staticmethod
    def download_file(url: str, target_path: str) -> None:
        """
        Downloads the file from the given URL to the given {target_path}
        """
        os.makedirs(os.path.dirname(target_path), exist_ok=True)
        try:
            response = requests.get(url, stream=True, timeout=60)
            if response.status_code != 200:
                log.error(f"Error downloading file '{url}': {response.status_code} {response.text}")
                raise SolidLSPException("Error downloading file.")
            with open(target_path, "wb") as f:
                shutil.copyfileobj(response.raw, f)
        except (requests.RequestException, OSError) as exc:
            log.error(f"Error downloading file '{url}': {exc}")
            raise SolidLSPException("Error downloading file.") from None

    @staticmethod
    def download_and_extract_archive(url: str, target_path: str, archive_type: str) -> None:
        """
        Downloads the archive from the given URL having format {archive_type} and extracts it to the given {target_path}
        """
        try:
            tmp_files = []
            tmp_file_name = str(PurePath(os.path.expanduser("~"), "solidlsp_tmp", uuid.uuid4().hex))
            tmp_files.append(tmp_file_name)
            os.makedirs(os.path.dirname(tmp_file_name), exist_ok=True)
            FileUtils.download_file(url, tmp_file_name)
            if archive_type in ["tar", "gztar", "bztar", "xztar"]:
                os.makedirs(target_path, exist_ok=True)
                shutil.unpack_archive(tmp_file_name, target_path, archive_type)
            elif archive_type == "zip":
                os.makedirs(target_path, exist_ok=True)
                with zipfile.ZipFile(tmp_file_name, "r") as zip_ref:
                    for zip_info in zip_ref.infolist():
                        extracted_path = zip_ref.extract(zip_info, target_path)
                        ZIP_SYSTEM_UNIX = 3  # zip file created on Unix system
                        if zip_info.create_system != ZIP_SYSTEM_UNIX:
                            continue
                        # extractall() does not preserve permissions
                        # see. https://github.com/python/cpython/issues/59999
                        attrs = (zip_info.external_attr >> 16) & 0o777
                        if attrs:
                            os.chmod(extracted_path, attrs)
            elif archive_type == "zip.gz":
                os.makedirs(target_path, exist_ok=True)
                tmp_file_name_ungzipped = tmp_file_name + ".zip"
                tmp_files.append(tmp_file_name_ungzipped)
                with gzip.open(tmp_file_name, "rb") as f_in, open(tmp_file_name_ungzipped, "wb") as f_out:
                    shutil.copyfileobj(f_in, f_out)
                shutil.unpack_archive(tmp_file_name_ungzipped, target_path, "zip")
            elif archive_type == "gz":
                with gzip.open(tmp_file_name, "rb") as f_in, open(target_path, "wb") as f_out:
                    shutil.copyfileobj(f_in, f_out)
            elif archive_type == "binary":
                # For single binary files, just move to target without extraction
                shutil.move(tmp_file_name, target_path)
            else:
                log.error(f"Unknown archive type '{archive_type}' for extraction")
                raise SolidLSPException(f"Unknown archive type '{archive_type}'")
        except (requests.RequestException, OSError, shutil.Error, zipfile.BadZipFile, gzip.BadGzipFile, ValueError) as exc:
            log.error(f"Error extracting archive '{tmp_file_name}' obtained from '{url}': {exc}")
            raise SolidLSPException("Error extracting archive.") from exc
        finally:
            for tmp_file_name in tmp_files:
                if os.path.exists(tmp_file_name):
                    Path.unlink(Path(tmp_file_name))


class PlatformId(str, Enum):
    WIN_x86 = "win-x86"
    WIN_x64 = "win-x64"
    WIN_arm64 = "win-arm64"
    OSX = "osx"
    OSX_x64 = "osx-x64"
    OSX_arm64 = "osx-arm64"
    LINUX_x86 = "linux-x86"
    LINUX_x64 = "linux-x64"
    LINUX_arm64 = "linux-arm64"
    LINUX_MUSL_x64 = "linux-musl-x64"
    LINUX_MUSL_arm64 = "linux-musl-arm64"

    def is_windows(self) -> bool:
        return self.value.startswith("win")


class DotnetVersion(str, Enum):
    V4 = "4"
    V6 = "6"
    V7 = "7"
    V8 = "8"
    V9 = "9"
    VMONO = "mono"


class PlatformUtils:
    """
    This class provides utilities for platform detection and identification.
    """

    @classmethod
    def get_platform_id(cls) -> PlatformId:
        """
        Returns the platform id for the current system
        """
        system = platform.system()
        machine = platform.machine()
        bitness = platform.architecture()[0]
        if system == "Windows" and machine == "":
            machine = cls._determine_windows_machine_type()
        system_map = {"Windows": "win", "Darwin": "osx", "Linux": "linux"}
        machine_map = {
            "AMD64": "x64",
            "x86_64": "x64",
            "i386": "x86",
            "i686": "x86",
            "aarch64": "arm64",
            "arm64": "arm64",
            "ARM64": "arm64",
        }
        if system in system_map and machine in machine_map:
            platform_id = system_map[system] + "-" + machine_map[machine]
            if system == "Linux" and bitness == "64bit":
                libc = platform.libc_ver()[0]
                if libc != "glibc":
                    # Format: linux-musl-arch (e.g., linux-musl-arm64)
                    platform_id = f"{system_map[system]}-{libc}-{machine_map[machine]}"
            return PlatformId(platform_id)
        else:
            raise SolidLSPException(f"Unknown platform: {system=}, {machine=}, {bitness=}")

    @staticmethod
    def _determine_windows_machine_type() -> str:
        import ctypes
        from ctypes import wintypes

        class SYSTEM_INFO(ctypes.Structure):
            class _U(ctypes.Union):
                class _S(ctypes.Structure):
                    _fields_ = [("wProcessorArchitecture", wintypes.WORD), ("wReserved", wintypes.WORD)]

                _fields_ = [("dwOemId", wintypes.DWORD), ("s", _S)]
                _anonymous_ = ("s",)

            _fields_ = [
                ("u", _U),
                ("dwPageSize", wintypes.DWORD),
                ("lpMinimumApplicationAddress", wintypes.LPVOID),
                ("lpMaximumApplicationAddress", wintypes.LPVOID),
                ("dwActiveProcessorMask", wintypes.LPVOID),
                ("dwNumberOfProcessors", wintypes.DWORD),
                ("dwProcessorType", wintypes.DWORD),
                ("dwAllocationGranularity", wintypes.DWORD),
                ("wProcessorLevel", wintypes.WORD),
                ("wProcessorRevision", wintypes.WORD),
            ]
            _anonymous_ = ("u",)

        sys_info = SYSTEM_INFO()
        ctypes.windll.kernel32.GetNativeSystemInfo(ctypes.byref(sys_info))  # type: ignore

        arch_map = {
            9: "AMD64",
            5: "ARM",
            12: "arm64",
            6: "Intel Itanium-based",
            0: "i386",
        }

        return arch_map.get(sys_info.wProcessorArchitecture, f"Unknown ({sys_info.wProcessorArchitecture})")

    @staticmethod
    def get_dotnet_version() -> DotnetVersion:
        """
        Returns the dotnet version for the current system
        """
        try:
            result = subprocess.run(["dotnet", "--list-runtimes"], capture_output=True, check=True)
            available_version_cmd_output = []
            for line in result.stdout.decode("utf-8").split("\n"):
                if line.startswith("Microsoft.NETCore.App"):
                    version_cmd_output = line.split(" ")[1]
                    available_version_cmd_output.append(version_cmd_output)

            if not available_version_cmd_output:
                raise SolidLSPException("dotnet not found on the system")

            # Check for supported versions in order of preference (latest first)
            for version_cmd_output in available_version_cmd_output:
                if version_cmd_output.startswith("9"):
                    return DotnetVersion.V9
                if version_cmd_output.startswith("8"):
                    return DotnetVersion.V8
                if version_cmd_output.startswith("7"):
                    return DotnetVersion.V7
                if version_cmd_output.startswith("6"):
                    return DotnetVersion.V6
                if version_cmd_output.startswith("4"):
                    return DotnetVersion.V4

            # If no supported version found, raise exception with all available versions
            raise SolidLSPException(
                f"No supported dotnet version found. Available versions: {', '.join(available_version_cmd_output)}. Supported versions: 4, 6, 7, 8"
            )
        except (FileNotFoundError, subprocess.CalledProcessError):
            try:
                result = subprocess.run(["mono", "--version"], capture_output=True, check=True)
                return DotnetVersion.VMONO
            except (FileNotFoundError, subprocess.CalledProcessError):
                raise SolidLSPException("dotnet or mono not found on the system")


class SymbolUtils:
    @staticmethod
    def symbol_tree_contains_name(roots: list[UnifiedSymbolInformation], name: str) -> bool:
        """
        Check if any symbol in the tree has a name matching the given name.
        """
        for symbol in roots:
            if symbol["name"] == name:
                return True
            if SymbolUtils.symbol_tree_contains_name(symbol["children"], name):
                return True
        return False



from __future__ import annotations

from enum import Enum, IntEnum, IntFlag
from typing import Literal, NotRequired, Union

from typing_extensions import TypedDict

class InlineValueRegistrationOptions(TypedDict):
    """Inline value options used during static or dynamic registration.

    @since 3.17.0
    """

    documentSelector: Union["DocumentSelector", None]
    """ A document selector to identify the scope of the registration. If set to null
    the document selector provided on the client side will be used. """
    id: NotRequired[str]
    """ The id used to register the request. The id can be used to deregister
    the request again. See also Registration#id. """


class InlayHintParams(TypedDict):
    """A parameter literal used in inlay hint requests.

    @since 3.17.0
    """

    textDocument: "TextDocumentIdentifier"
    """ The text document. """
    range: "Range"
    """ The document range for which inlay hints should be computed. """
    workDoneToken: NotRequired["ProgressToken"]
    """ An optional token that a server can use to report work done progress. """


class InlayHint(TypedDict):
    """Inlay hint information.

    @since 3.17.0
    """

    position: "Position"
    """ The position of this hint. """
    label: str | list["InlayHintLabelPart"]
    """ The label of this hint. A human readable string or an array of
    InlayHintLabelPart label parts.

    *Note* that neither the string nor the label part can be empty. """
    kind: NotRequired["InlayHintKind"]
    """ The kind of this hint. Can be omitted in which case the client
    should fall back to a reasonable default. """
    textEdits: NotRequired[list["TextEdit"]]
    """ Optional text edits that are performed when accepting this inlay hint.

    *Note* that edits are expected to change the document so that the inlay
    hint (or its nearest variant) is now part of the document and the inlay
    hint itself is now obsolete. """
    tooltip: NotRequired[Union[str, "MarkupContent"]]
    """ The tooltip text when you hover over this item. """
    paddingLeft: NotRequired[bool]
    """ Render padding before the hint.

    Note: Padding should use the editor's background color, not the
    background color of the hint itself. That means padding can be used
    to visually align/separate an inlay hint. """
    paddingRight: NotRequired[bool]
    """ Render padding after the hint.

    Note: Padding should use the editor's background color, not the
    background color of the hint itself. That means padding can be used
    to visually align/separate an inlay hint. """
    data: NotRequired["LSPAny"]
    """ A data entry field that is preserved on an inlay hint between
    a `textDocument/inlayHint` and a `inlayHint/resolve` request. """


class InlayHintRegistrationOptions(TypedDict):
    """Inlay hint options used during static or dynamic registration.

    @since 3.17.0
    """

    resolveProvider: NotRequired[bool]
    """ The server provides support to resolve additional
    information for an inlay hint item. """
    documentSelector: Union["DocumentSelector", None]
    """ A document selector to identify the scope of the registration. If set to null
    the document selector provided on the client side will be used. """
    id: NotRequired[str]
    """ The id used to register the request. The id can be used to deregister
    the request again. See also Registration#id. """


class DocumentDiagnosticParams(TypedDict):
    """Parameters of the document diagnostic request.

    @since 3.17.0
    """

    textDocument: "TextDocumentIdentifier"
    """ The text document. """
    identifier: NotRequired[str]
    """ The additional identifier  provided during registration. """
    previousResultId: NotRequired[str]
    """ The result id of a previous response if provided. """
    workDoneToken: NotRequired["ProgressToken"]
    """ An optional token that a server can use to report work done progress. """
    partialResultToken: NotRequired["ProgressToken"]
    """ An optional token that a server can use to report partial results (e.g. streaming) to
    the client. """


class DocumentDiagnosticReportPartialResult(TypedDict):
    """A partial result for a document diagnostic report.

    @since 3.17.0
    """

    relatedDocuments: dict[
        "DocumentUri",
        Union["FullDocumentDiagnosticReport", "UnchangedDocumentDiagnosticReport"],
    ]


class DiagnosticServerCancellationData(TypedDict):
    """Cancellation data returned from a diagnostic request.

    @since 3.17.0
    """

    retriggerRequest: bool


class DiagnosticRegistrationOptions(TypedDict):
    """Diagnostic registration options.

    @since 3.17.0
    """

    documentSelector: Union["DocumentSelector", None]
    """ A document selector to identify the scope of the registration. If set to null
    the document selector provided on the client side will be used. """
    identifier: NotRequired[str]
    """ An optional identifier under which the diagnostics are
    managed by the client. """
    interFileDependencies: bool
    """ Whether the language has inter file dependencies meaning that
    editing code in one file can result in a different diagnostic
    set in another file. Inter file dependencies are common for
    most programming languages and typically uncommon for linters. """
    workspaceDiagnostics: bool
    """ The server provides support for workspace diagnostics as well. """
    id: NotRequired[str]
    """ The id used to register the request. The id can be used to deregister
    the request again. See also Registration#id. """


class WorkspaceDiagnosticParams(TypedDict):
    """Parameters of the workspace diagnostic request.

    @since 3.17.0
    """

    identifier: NotRequired[str]
    """ The additional identifier provided during registration. """
    previousResultIds: list["PreviousResultId"]
    """ The currently known diagnostic reports with their
    previous result ids. """
    workDoneToken: NotRequired["ProgressToken"]
    """ An optional token that a server can use to report work done progress. """
    partialResultToken: NotRequired["ProgressToken"]
    """ An optional token that a server can use to report partial results (e.g. streaming) to
    the client. """


class WorkspaceDiagnosticReport(TypedDict):
    """A workspace diagnostic report.

    @since 3.17.0
    """

    items: list["WorkspaceDocumentDiagnosticReport"]


class WorkspaceDiagnosticReportPartialResult(TypedDict):
    """A partial result for a workspace diagnostic report.

    @since 3.17.0
    """

    items: list["WorkspaceDocumentDiagnosticReport"]


class DidOpenNotebookDocumentParams(TypedDict):
    """The params sent in an open notebook document notification.

    @since 3.17.0
    """

    notebookDocument: "NotebookDocument"
    """ The notebook document that got opened. """
    cellTextDocuments: list["TextDocumentItem"]
    """ The text documents that represent the content
    of a notebook cell. """


class DidChangeNotebookDocumentParams(TypedDict):
    """The params sent in a change notebook document notification.

    @since 3.17.0
    """

    notebookDocument: "VersionedNotebookDocumentIdentifier"
    """ The notebook document that did change. The version number points
    to the version after all provided changes have been applied. If
    only the text document content of a cell changes the notebook version
    doesn't necessarily have to change. """
    change: "NotebookDocumentChangeEvent"
    """ The actual changes to the notebook document.

    The changes describe single state changes to the notebook document.
    So if there are two changes c1 (at array index 0) and c2 (at array
    index 1) for a notebook in state S then c1 moves the notebook from
    S to S' and c2 from S' to S''. So c1 is computed on the state S and
    c2 is computed on the state S'.

    To mirror the content of a notebook using change events use the following approach:
    - start with the same initial content
    - apply the 'notebookDocument/didChange' notifications in the order you receive them.
    - apply the `NotebookChangeEvent`s in a single notification in the order
      you receive them. """


class DidSaveNotebookDocumentParams(TypedDict):
    """The params sent in a save notebook document notification.

    @since 3.17.0
    """

    notebookDocument: "NotebookDocumentIdentifier"
    """ The notebook document that got saved. """


class DidCloseNotebookDocumentParams(TypedDict):
    """The params sent in a close notebook document notification.

    @since 3.17.0
    """

    notebookDocument: "NotebookDocumentIdentifier"
    """ The notebook document that got closed. """
    cellTextDocuments: list["TextDocumentIdentifier"]
    """ The text documents that represent the content
    of a notebook cell that got closed. """


class RegistrationParams(TypedDict):
    registrations: list["Registration"]


class UnregistrationParams(TypedDict):
    unregisterations: list["Unregistration"]


class InitializeParams(TypedDict):
    processId: int | None
    """ The process Id of the parent process that started
    the server.

    Is `null` if the process has not been started by another process.
    If the parent process is not alive then the server should exit. """
    clientInfo: NotRequired["___InitializeParams_clientInfo_Type_1"]
    """ Information about the client

    @since 3.15.0 """
    locale: NotRequired[str]
    """ The locale the client is currently showing the user interface
    in. This must not necessarily be the locale of the operating
    system.

    Uses IETF language tags as the value's syntax
    (See https://en.wikipedia.org/wiki/IETF_language_tag)

    @since 3.16.0 """
    rootPath: NotRequired[str | None]
    """ The rootPath of the workspace. Is null
    if no folder is open.

    @deprecated in favour of rootUri. """
    rootUri: Union["DocumentUri", None]
    """ The rootUri of the workspace. Is null if no
    folder is open. If both `rootPath` and `rootUri` are set
    `rootUri` wins.

    @deprecated in favour of workspaceFolders. """
    capabilities: "ClientCapabilities"
    """ The capabilities provided by the client (editor or tool) """
    initializationOptions: NotRequired["LSPAny"]
    """ User provided initialization options. """
    trace: NotRequired["TraceValues"]
    """ The initial trace setting. If omitted trace is disabled ('off'). """
    workspaceFolders: NotRequired[list["WorkspaceFolder"] | None]
    """ The workspace folders configured in the client when the server starts.

    This property is only available if the client supports workspace folders.
    It can be `null` if the client supports workspace folders but none are
    configured.

    @since 3.6.0 """


class InitializeResult(TypedDict):
    """The result returned from an initialize request."""

    capabilities: "ServerCapabilities"
    """ The capabilities the language server provides. """
    serverInfo: NotRequired["__InitializeResult_serverInfo_Type_1"]
    """ Information about the server.

    @since 3.15.0 """


class InitializeError(TypedDict):
    """The data type of the ResponseError if the
    initialize request fails.
    """

    retry: bool
    """ Indicates whether the client execute the following retry logic:
    (1) show the message provided by the ResponseError to the user
    (2) user selects retry or cancel
    (3) if user selected retry the initialize method is sent again. """


class InitializedParams(TypedDict):
    ...


class DidChangeConfigurationParams(TypedDict):
    """The parameters of a change configuration notification."""

    settings: "LSPAny"
    """ The actual changed settings """


class DidChangeConfigurationRegistrationOptions(TypedDict):
    section: NotRequired[str | list[str]]


class ShowMessageParams(TypedDict):
    """The parameters of a notification message."""

    type: "MessageType"
    """ The message type. See {@link MessageType} """
    message: str
    """ The actual message. """


class ShowMessageRequestParams(TypedDict):
    type: "MessageType"
    """ The message type. See {@link MessageType} """
    message: str
    """ The actual message. """
    actions: NotRequired[list["MessageActionItem"]]
    """ The message action items to present. """


class MessageActionItem(TypedDict):
    title: str
    """ A short title like 'Retry', 'Open Log' etc. """


class LogMessageParams(TypedDict):
    """The log message parameters."""

    type: "MessageType"
    """ The message type. See {@link MessageType} """
    message: str
    """ The actual message. """


class DidOpenTextDocumentParams(TypedDict):
    """The parameters sent in an open text document notification"""

    textDocument: "TextDocumentItem"
    """ The document that was opened. """


class DidChangeTextDocumentParams(TypedDict):
    """The change text document notification's parameters."""

    textDocument: "VersionedTextDocumentIdentifier"
    """ The document that did change. The version number points
    to the version after all provided content changes have
    been applied. """
    contentChanges: list["TextDocumentContentChangeEvent"]
    """ The actual content changes. The content changes describe single state changes
    to the document. So if there are two content changes c1 (at array index 0) and
    c2 (at array index 1) for a document in state S then c1 moves the document from
    S to S' and c2 from S' to S''. So c1 is computed on the state S and c2 is computed
    on the state S'.

    To mirror the content of a document using change events use the following approach:
    - start with the same initial content
    - apply the 'textDocument/didChange' notifications in the order you receive them.
    - apply the `TextDocumentContentChangeEvent`s in a single notification in the order
      you receive them. """


class TextDocumentChangeRegistrationOptions(TypedDict):
    """Describe options to be used when registered for text document change events."""

    syncKind: "TextDocumentSyncKind"
    """ How documents are synced to the server. """
    documentSelector: Union["DocumentSelector", None]
    """ A document selector to identify the scope of the registration. If set to null
    the document selector provided on the client side will be used. """


class DidCloseTextDocumentParams(TypedDict):
    """The parameters sent in a close text document notification"""

    textDocument: "TextDocumentIdentifier"
    """ The document that was closed. """


class DidSaveTextDocumentParams(TypedDict):
    """The parameters sent in a save text document notification"""

    textDocument: "TextDocumentIdentifier"
    """ The document that was saved. """
    text: NotRequired[str]
    """ Optional the content when saved. Depends on the includeText value
    when the save notification was requested. """


class TextDocumentSaveRegistrationOptions(TypedDict):
    """Save registration options."""

    documentSelector: Union["DocumentSelector", None]
    """ A document selector to identify the scope of the registration. If set to null
    the document selector provided on the client side will be used. """
    includeText: NotRequired[bool]
    """ The client is supposed to include the content on save. """


class WillSaveTextDocumentParams(TypedDict):
    """The parameters sent in a will save text document notification."""

    textDocument: "TextDocumentIdentifier"
    """ The document that will be saved. """
    reason: "TextDocumentSaveReason"
    """ The 'TextDocumentSaveReason'. """


class TextEdit(TypedDict):
    """A text edit applicable to a text document."""

    range: "Range"
    """ The range of the text document to be manipulated. To insert
    text into a document create a range where start === end. """
    newText: str
    """ The string to be inserted. For delete operations use an
    empty string. """


class DidChangeWatchedFilesParams(TypedDict):
    """The watched files change notification's parameters."""

    changes: list["FileEvent"]
    """ The actual file events. """


class DidChangeWatchedFilesRegistrationOptions(TypedDict):
    """Describe options to be used when registered for text document change events."""

    watchers: list["FileSystemWatcher"]
    """ The watchers to register. """


class PublishDiagnosticsParams(TypedDict):
    """The publish diagnostic notification's parameters."""

    uri: "DocumentUri"
    """ The URI for which diagnostic information is reported. """
    version: NotRequired[int]
    """ Optional the version number of the document the diagnostics are published for.

    @since 3.15.0 """
    diagnostics: list["Diagnostic"]
    """ An array of diagnostic information items. """


class CompletionParams(TypedDict):
    """Completion parameters"""

    context: NotRequired["CompletionContext"]
    """ The completion context. This is only available it the client specifies
    to send this using the client capability `textDocument.completion.contextSupport === true` """
    textDocument: "TextDocumentIdentifier"
    """ The text document. """
    position: "Position"
    """ The position inside the text document. """
    workDoneToken: NotRequired["ProgressToken"]
    """ An optional token that a server can use to report work done progress. """
    partialResultToken: NotRequired["ProgressToken"]
    """ An optional token that a server can use to report partial results (e.g. streaming) to
    the client. """


class CompletionItem(TypedDict):
    """A completion item represents a text snippet that is
    proposed to complete text that is being typed.
    """

    label: str
    """ The label of this completion item.

    The label property is also by default the text that
    is inserted when selecting this completion.

    If label details are provided the label itself should
    be an unqualified name of the completion item. """
    labelDetails: NotRequired["CompletionItemLabelDetails"]
    """ Additional details for the label

    @since 3.17.0 """
    kind: NotRequired["CompletionItemKind"]
    """ The kind of this completion item. Based of the kind
    an icon is chosen by the editor. """
    tags: NotRequired[list["CompletionItemTag"]]
    """ Tags for this completion item.

    @since 3.15.0 """
    detail: NotRequired[str]
    """ A human-readable string with additional information
    about this item, like type or symbol information. """
    documentation: NotRequired[Union[str, "MarkupContent"]]
    """ A human-readable string that represents a doc-comment. """
    deprecated: NotRequired[bool]
    """ Indicates if this item is deprecated.
    @deprecated Use `tags` instead. """
    preselect: NotRequired[bool]
    """ Select this item when showing.

    *Note* that only one completion item can be selected and that the
    tool / client decides which item that is. The rule is that the *first*
    item of those that match best is selected. """
    sortText: NotRequired[str]
    """ A string that should be used when comparing this item
    with other items. When `falsy` the {@link CompletionItem.label label}
    is used. """
    filterText: NotRequired[str]
    """ A string that should be used when filtering a set of
    completion items. When `falsy` the {@link CompletionItem.label label}
    is used. """
    insertText: NotRequired[str]
    """ A string that should be inserted into a document when selecting
    this completion. When `falsy` the {@link CompletionItem.label label}
    is used.

    The `insertText` is subject to interpretation by the client side.
    Some tools might not take the string literally. For example
    VS Code when code complete is requested in this example
    `con<cursor position>` and a completion item with an `insertText` of
    `console` is provided it will only insert `sole`. Therefore it is
    recommended to use `textEdit` instead since it avoids additional client
    side interpretation. """
    insertTextFormat: NotRequired["InsertTextFormat"]
    """ The format of the insert text. The format applies to both the
    `insertText` property and the `newText` property of a provided
    `textEdit`. If omitted defaults to `InsertTextFormat.PlainText`.

    Please note that the insertTextFormat doesn't apply to
    `additionalTextEdits`. """
    insertTextMode: NotRequired["InsertTextMode"]
    """ How whitespace and indentation is handled during completion
    item insertion. If not provided the clients default value depends on
    the `textDocument.completion.insertTextMode` client capability.

    @since 3.16.0 """
    textEdit: NotRequired[Union["TextEdit", "InsertReplaceEdit"]]
    """ An {@link TextEdit edit} which is applied to a document when selecting
    this completion. When an edit is provided the value of
    {@link CompletionItem.insertText insertText} is ignored.

    Most editors support two different operations when accepting a completion
    item. One is to insert a completion text and the other is to replace an
    existing text with a completion text. Since this can usually not be
    predetermined by a server it can report both ranges. Clients need to
    signal support for `InsertReplaceEdits` via the
    `textDocument.completion.insertReplaceSupport` client capability
    property.

    *Note 1:* The text edit's range as well as both ranges from an insert
    replace edit must be a [single line] and they must contain the position
    at which completion has been requested.
    *Note 2:* If an `InsertReplaceEdit` is returned the edit's insert range
    must be a prefix of the edit's replace range, that means it must be
    contained and starting at the same position.

    @since 3.16.0 additional type `InsertReplaceEdit` """
    textEditText: NotRequired[str]
    """ The edit text used if the completion item is part of a CompletionList and
    CompletionList defines an item default for the text edit range.

    Clients will only honor this property if they opt into completion list
    item defaults using the capability `completionList.itemDefaults`.

    If not provided and a list's default range is provided the label
    property is used as a text.

    @since 3.17.0 """
    additionalTextEdits: NotRequired[list["TextEdit"]]
    """ An optional array of additional {@link TextEdit text edits} that are applied when
    selecting this completion. Edits must not overlap (including the same insert position)
    with the main {@link CompletionItem.textEdit edit} nor with themselves.

    Additional text edits should be used to change text unrelated to the current cursor position
    (for example adding an import statement at the top of the file if the completion item will
    insert an unqualified type). """
    commitCharacters: NotRequired[list[str]]
    """ An optional set of characters that when pressed while this completion is active will accept it first and
    then type that character. *Note* that all commit characters should have `length=1` and that superfluous
    characters will be ignored. """
    command: NotRequired["Command"]
    """ An optional {@link Command command} that is executed *after* inserting this completion. *Note* that
    additional modifications to the current document should be described with the
    {@link CompletionItem.additionalTextEdits additionalTextEdits}-property. """
    data: NotRequired["LSPAny"]
    """ A data entry field that is preserved on a completion item between a
    {@link CompletionRequest} and a {@link CompletionResolveRequest}. """


class CompletionList(TypedDict):
    """Represents a collection of {@link CompletionItem completion items} to be presented
    in the editor.
    """

    isIncomplete: bool
    """ This list it not complete. Further typing results in recomputing this list.

    Recomputed lists have all their items replaced (not appended) in the
    incomplete completion sessions. """
    itemDefaults: NotRequired["__CompletionList_itemDefaults_Type_1"]
    """ In many cases the items of an actual completion result share the same
    value for properties like `commitCharacters` or the range of a text
    edit. A completion list can therefore define item defaults which will
    be used if a completion item itself doesn't specify the value.

    If a completion list specifies a default value and a completion item
    also specifies a corresponding value the one from the item is used.

    Servers are only allowed to return default values if the client
    signals support for this via the `completionList.itemDefaults`
    capability.

    @since 3.17.0 """
    items: list["CompletionItem"]
    """ The completion items. """


class CompletionRegistrationOptions(TypedDict):
    """Registration options for a {@link CompletionRequest}."""

    documentSelector: Union["DocumentSelector", None]
    """ A document selector to identify the scope of the registration. If set to null
    the document selector provided on the client side will be used. """
    triggerCharacters: NotRequired[list[str]]
    """ Most tools trigger completion request automatically without explicitly requesting
    it using a keyboard shortcut (e.g. Ctrl+Space). Typically they do so when the user
    starts to type an identifier. For example if the user types `c` in a JavaScript file
    code complete will automatically pop up present `console` besides others as a
    completion item. Characters that make up identifiers don't need to be listed here.

    If code complete should automatically be trigger on characters not being valid inside
    an identifier (for example `.` in JavaScript) list them in `triggerCharacters`. """
    allCommitCharacters: NotRequired[list[str]]
    """ The list of all possible characters that commit a completion. This field can be used
    if clients don't support individual commit characters per completion item. See
    `ClientCapabilities.textDocument.completion.completionItem.commitCharactersSupport`

    If a server provides both `allCommitCharacters` and commit characters on an individual
    completion item the ones on the completion item win.

    @since 3.2.0 """
    resolveProvider: NotRequired[bool]
    """ The server provides support to resolve additional
    information for a completion item. """
    completionItem: NotRequired["__CompletionOptions_completionItem_Type_1"]
    """ The server supports the following `CompletionItem` specific
    capabilities.

    @since 3.17.0 """


class HoverParams(TypedDict):
    """Parameters for a {@link HoverRequest}."""

    textDocument: "TextDocumentIdentifier"
    """ The text document. """
    position: "Position"
    """ The position inside the text document. """
    workDoneToken: NotRequired["ProgressToken"]
    """ An optional token that a server can use to report work done progress. """


class Hover(TypedDict):
    """The result of a hover request."""

    contents: Union["MarkupContent", "MarkedString", list["MarkedString"]]
    """ The hover's content """
    range: NotRequired["Range"]
    """ An optional range inside the text document that is used to
    visualize the hover, e.g. by changing the background color. """


class HoverRegistrationOptions(TypedDict):
    """Registration options for a {@link HoverRequest}."""

    documentSelector: Union["DocumentSelector", None]
    """ A document selector to identify the scope of the registration. If set to null
    the document selector provided on the client side will be used. """


class SignatureHelpParams(TypedDict):
    """Parameters for a {@link SignatureHelpRequest}."""

    context: NotRequired["SignatureHelpContext"]
    """ The signature help context. This is only available if the client specifies
    to send this using the client capability `textDocument.signatureHelp.contextSupport === true`

    @since 3.15.0 """
    textDocument: "TextDocumentIdentifier"
    """ The text document. """
    position: "Position"
    """ The position inside the text document. """
    workDoneToken: NotRequired["ProgressToken"]
    """ An optional token that a server can use to report work done progress. """




"""파일 수집 L1 저장소를 구현한다."""

from __future__ import annotations

from pathlib import Path

from sari.core.models import CollectedFileL1DTO, EnrichStateUpdateDTO, FileListItemDTO
from sari.db.row_mapper import row_bool, row_int, row_str
from sari.db.schema import connect


class FileCollectionRepository:
    """L1 파일 메타데이터 영속화를 담당한다."""

    def __init__(self, db_path: Path) -> None:
        """저장소에 사용할 DB 경로를 저장한다."""
        self._db_path = db_path

    def upsert_file(self, file_row: CollectedFileL1DTO) -> None:
        """L1 파일 메타데이터를 업서트한다."""
        with connect(self._db_path) as conn:
            conn.execute(
                """
                INSERT INTO collected_files_l1(
                    repo_root, relative_path, absolute_path, repo_label, mtime_ns, size_bytes,
                    content_hash, is_deleted, last_seen_at, updated_at, enrich_state
                )
                VALUES(
                    :repo_root, :relative_path, :absolute_path, :repo_label, :mtime_ns, :size_bytes,
                    :content_hash, :is_deleted, :last_seen_at, :updated_at, :enrich_state
                )
                ON CONFLICT(repo_root, relative_path) DO UPDATE SET
                    absolute_path = excluded.absolute_path,
                    repo_label = excluded.repo_label,
                    mtime_ns = excluded.mtime_ns,
                    size_bytes = excluded.size_bytes,
                    content_hash = excluded.content_hash,
                    is_deleted = excluded.is_deleted,
                    last_seen_at = excluded.last_seen_at,
                    updated_at = excluded.updated_at,
                    enrich_state = excluded.enrich_state
                """,
                file_row.to_sql_params(),
            )
            conn.commit()

    def upsert_files_many(self, file_rows: list[CollectedFileL1DTO]) -> None:
        """L1 파일 메타데이터를 배치 업서트한다."""
        if len(file_rows) == 0:
            return
        with connect(self._db_path) as conn:
            for file_row in file_rows:
                conn.execute(
                    """
                    INSERT INTO collected_files_l1(
                        repo_root, relative_path, absolute_path, repo_label, mtime_ns, size_bytes,
                        content_hash, is_deleted, last_seen_at, updated_at, enrich_state
                    )
                    VALUES(
                        :repo_root, :relative_path, :absolute_path, :repo_label, :mtime_ns, :size_bytes,
                        :content_hash, :is_deleted, :last_seen_at, :updated_at, :enrich_state
                    )
                    ON CONFLICT(repo_root, relative_path) DO UPDATE SET
                        absolute_path = excluded.absolute_path,
                        repo_label = excluded.repo_label,
                        mtime_ns = excluded.mtime_ns,
                        size_bytes = excluded.size_bytes,
                        content_hash = excluded.content_hash,
                        is_deleted = excluded.is_deleted,
                        last_seen_at = excluded.last_seen_at,
                        updated_at = excluded.updated_at,
                        enrich_state = excluded.enrich_state
                    """,
                    file_row.to_sql_params(),
                )
            conn.commit()

    def list_files(self, repo_root: str, limit: int, prefix: str | None = None) -> list[FileListItemDTO]:
        """활성 파일 목록을 조회한다."""
        where_prefix = ""
        params: dict[str, object] = {"repo_root": repo_root, "limit": limit}
        if prefix is not None and prefix.strip() != "":
            where_prefix = "AND relative_path LIKE :prefix"
            params["prefix"] = f"{prefix.strip()}%"

        with connect(self._db_path) as conn:
            rows = conn.execute(
                f"""
                SELECT repo_root, relative_path, size_bytes, mtime_ns, content_hash, enrich_state
                FROM collected_files_l1
                WHERE repo_root = :repo_root
                  AND is_deleted = 0
                  {where_prefix}
                ORDER BY relative_path ASC
                LIMIT :limit
                """,
                params,
            ).fetchall()

        items: list[FileListItemDTO] = []
        for row in rows:
            items.append(
                FileListItemDTO(
                    repo=row_str(row, "repo_root"),
                    relative_path=row_str(row, "relative_path"),
                    size_bytes=row_int(row, "size_bytes"),
                    mtime_ns=row_int(row, "mtime_ns"),
                    content_hash=row_str(row, "content_hash"),
                    enrich_state=row_str(row, "enrich_state"),
                )
            )
        return items

    def get_file(self, repo_root: str, relative_path: str) -> CollectedFileL1DTO | None:
        """단일 파일 메타데이터를 조회한다."""
        with connect(self._db_path) as conn:
            row = conn.execute(
                """
                SELECT repo_root, relative_path, absolute_path, repo_label, mtime_ns, size_bytes, content_hash,
                       is_deleted, last_seen_at, updated_at, enrich_state
                FROM collected_files_l1
                WHERE repo_root = :repo_root
                  AND relative_path = :relative_path
                """,
                {"repo_root": repo_root, "relative_path": relative_path},
            ).fetchone()
        if row is None:
            return None
        return CollectedFileL1DTO(
            repo_root=row_str(row, "repo_root"),
            relative_path=row_str(row, "relative_path"),
            absolute_path=row_str(row, "absolute_path"),
            repo_label=row_str(row, "repo_label"),
            mtime_ns=row_int(row, "mtime_ns"),
            size_bytes=row_int(row, "size_bytes"),
            content_hash=row_str(row, "content_hash"),
            is_deleted=row_bool(row, "is_deleted"),
            last_seen_at=row_str(row, "last_seen_at"),
            updated_at=row_str(row, "updated_at"),
            enrich_state=row_str(row, "enrich_state"),
        )

    def mark_missing_as_deleted(self, repo_root: str, seen_relative_paths: list[str], updated_at: str, scan_started_at: str) -> int:
        """스캔 시작 시각 이전 last_seen 항목 중 누락 파일을 삭제 상태로 전환한다."""
        with connect(self._db_path) as conn:
            if len(seen_relative_paths) == 0:
                cur = conn.execute(
                    """
                    UPDATE collected_files_l1
                    SET is_deleted = 1,
                        updated_at = :updated_at,
                        enrich_state = 'DELETED'
                    WHERE repo_root = :repo_root
                      AND is_deleted = 0
                      AND last_seen_at < :scan_started_at
                    """,
                    {"repo_root": repo_root, "updated_at": updated_at, "scan_started_at": scan_started_at},
                )
                conn.commit()
                return int(cur.rowcount if cur.rowcount is not None else 0)

            placeholders: list[str] = []
            params: dict[str, object] = {"repo_root": repo_root, "updated_at": updated_at, "scan_started_at": scan_started_at}
            for index, value in enumerate(seen_relative_paths):
                key = f"seen_{index}"
                placeholders.append(f":{key}")
                params[key] = value

            cur = conn.execute(
                f"""
                UPDATE collected_files_l1
                SET is_deleted = 1,
                    updated_at = :updated_at,
                    enrich_state = 'DELETED'
                WHERE repo_root = :repo_root
                  AND is_deleted = 0
                  AND last_seen_at < :scan_started_at
                  AND relative_path NOT IN ({", ".join(placeholders)})
                """,
                params,
            )
            conn.commit()
            return int(cur.rowcount if cur.rowcount is not None else 0)

    def update_enrich_state(self, repo_root: str, relative_path: str, enrich_state: str, updated_at: str) -> None:
        """파일 보강 상태를 갱신한다."""
        with connect(self._db_path) as conn:
            conn.execute(
                """
                UPDATE collected_files_l1
                SET enrich_state = :enrich_state,
                    updated_at = :updated_at
                WHERE repo_root = :repo_root
                  AND relative_path = :relative_path
                """,
                {
                    "repo_root": repo_root,
                    "relative_path": relative_path,
                    "enrich_state": enrich_state,
                    "updated_at": updated_at,
                },
            )
            conn.commit()

    def update_enrich_state_many(self, updates: list[EnrichStateUpdateDTO]) -> None:
        """파일 보강 상태를 배치로 갱신한다."""
        if len(updates) == 0:
            return
        with connect(self._db_path) as conn:
            for item in updates:
                conn.execute(
                    """
                    UPDATE collected_files_l1
                    SET enrich_state = :enrich_state,
                        updated_at = :updated_at
                    WHERE repo_root = :repo_root
                      AND relative_path = :relative_path
                    """,
                    item.to_sql_params(),
                )
            conn.commit()

    def mark_deleted(self, repo_root: str, relative_path: str, updated_at: str) -> None:
        """특정 파일을 삭제 상태로 전환한다."""
        with connect(self._db_path) as conn:
            conn.execute(
                """
                UPDATE collected_files_l1
                SET is_deleted = 1,
                    updated_at = :updated_at,
                    enrich_state = 'DELETED'
                WHERE repo_root = :repo_root
                  AND relative_path = :relative_path
                """,
                {
                    "repo_root": repo_root,
                    "relative_path": relative_path,
                    "updated_at": updated_at,
                },
            )
            conn.commit()

    def get_repo_stats(self) -> list[dict[str, object]]:
        """저장소별 파일 수 통계를 조회한다."""
        with connect(self._db_path) as conn:
            rows = conn.execute(
                """
                SELECT repo_root, COUNT(relative_path) AS file_count
                FROM collected_files_l1
                WHERE is_deleted = 0
                GROUP BY repo_root
                ORDER BY repo_root ASC
                """
            ).fetchall()
        stats: list[dict[str, object]] = []
        for row in rows:
            stats.append({"repo": row_str(row, "repo_root"), "file_count": row_int(row, "file_count")})
        return stats

    def get_enrich_state_counts(self) -> dict[str, int]:
        """활성 파일의 enrich_state 분포를 조회한다."""
        with connect(self._db_path) as conn:
            rows = conn.execute(
                """
                SELECT enrich_state, COUNT(relative_path) AS file_count
                FROM collected_files_l1
                WHERE is_deleted = 0
                GROUP BY enrich_state
                """
            ).fetchall()
        counts: dict[str, int] = {}
        for row in rows:
            counts[row_str(row, "enrich_state")] = row_int(row, "file_count")
        return counts



"""Alembic 마이그레이션 실행 유틸리티를 제공한다."""

from __future__ import annotations

import os
from pathlib import Path
import sqlite3

HEAD_VERSION = "20260218_0004"
BASELINE_VERSION = "20260216_0001"


def ensure_migrated(db_path: Path) -> None:
    """DB를 Alembic head 리비전까지 업그레이드한다."""
    if os.getenv("SARI_DISABLE_ALEMBIC_AUTO", "").strip().lower() in {"1", "true", "yes", "on"}:
        return
    try:
        from alembic import command
        from alembic.config import Config
    except (ImportError, ModuleNotFoundError):
        _fallback_upgrade_sqlite(db_path)
        return
    project_root = Path(__file__).resolve().parents[3]
    alembic_ini = project_root / "alembic.ini"
    alembic_dir = project_root / "alembic"
    if not alembic_ini.exists() or not alembic_dir.exists():
        _fallback_upgrade_sqlite(db_path)
        return
    config = Config(str(alembic_ini))
    config.set_main_option("script_location", str(alembic_dir))
    config.set_main_option("sqlalchemy.url", f"sqlite:///{db_path}")
    # CLI JSON stdout 오염 방지를 위해 Alembic 기본 로거 설정을 비활성화한다.
    config.attributes["configure_logger"] = False
    command.upgrade(config, "head")


def _connect(db_path: Path) -> sqlite3.Connection:
    """SQLite 연결을 생성한다."""
    conn = sqlite3.connect(str(db_path))
    conn.row_factory = sqlite3.Row
    conn.execute("PRAGMA synchronous=NORMAL")
    conn.execute("PRAGMA busy_timeout=5000")
    return conn


def _table_columns(conn: sqlite3.Connection, table_name: str) -> set[str]:
    """테이블 컬럼 집합을 조회한다."""
    rows = conn.execute(f"PRAGMA table_info({table_name})").fetchall()
    return {str(row["name"]) for row in rows}


def _ensure_version_table(conn: sqlite3.Connection) -> str:
    """alembic_version 테이블을 보장하고 현재 버전을 반환한다."""
    conn.execute(
        """
        CREATE TABLE IF NOT EXISTS alembic_version (
            version_num VARCHAR(32) NOT NULL PRIMARY KEY
        )
        """
    )
    row = conn.execute("SELECT version_num FROM alembic_version LIMIT 1").fetchone()
    if row is None:
        conn.execute("INSERT INTO alembic_version(version_num) VALUES(:version_num)", {"version_num": BASELINE_VERSION})
        return BASELINE_VERSION
    return str(row["version_num"])


def _set_version(conn: sqlite3.Connection, version_num: str) -> None:
    """alembic_version 값을 갱신한다."""
    conn.execute("UPDATE alembic_version SET version_num = :version_num", {"version_num": version_num})


def _fallback_upgrade_sqlite(db_path: Path) -> None:
    """alembic 미설치 환경에서 head 스키마를 보장한다."""
    with _connect(db_path) as conn:
        current_version = _ensure_version_table(conn)
        if current_version < "20260217_0002":
            _fallback_upgrade_0002(conn)
            _set_version(conn, "20260217_0002")
            current_version = "20260217_0002"
        if current_version < "20260217_0003":
            _fallback_upgrade_0003(conn)
            _set_version(conn, "20260217_0003")
            current_version = "20260217_0003"
        if current_version < "20260218_0004":
            _fallback_upgrade_0004(conn)
            _set_version(conn, "20260218_0004")
        conn.commit()


def _fallback_upgrade_0002(conn: sqlite3.Connection) -> None:
    """0002 리비전 컬럼을 적용한다."""
    daemon_cols = _table_columns(conn, "daemon_runtime")
    if "last_heartbeat_at" not in daemon_cols:
        conn.execute("ALTER TABLE daemon_runtime ADD COLUMN last_heartbeat_at TEXT NULL")
    if "last_exit_reason" not in daemon_cols:
        conn.execute("ALTER TABLE daemon_runtime ADD COLUMN last_exit_reason TEXT NULL")
    conn.execute(
        """
        UPDATE daemon_runtime
        SET last_heartbeat_at = COALESCE(last_heartbeat_at, started_at)
        WHERE singleton_key = 'default'
        """
    )
    queue_cols = _table_columns(conn, "file_enrich_queue")
    if "priority" not in queue_cols:
        conn.execute("ALTER TABLE file_enrich_queue ADD COLUMN priority INTEGER NOT NULL DEFAULT 30")
    if "enqueue_source" not in queue_cols:
        conn.execute("ALTER TABLE file_enrich_queue ADD COLUMN enqueue_source TEXT NOT NULL DEFAULT 'scan'")
    policy_cols = _table_columns(conn, "pipeline_policy")
    if "bootstrap_mode_enabled" not in policy_cols:
        conn.execute("ALTER TABLE pipeline_policy ADD COLUMN bootstrap_mode_enabled INTEGER NOT NULL DEFAULT 0")
    if "bootstrap_l3_worker_count" not in policy_cols:
        conn.execute("ALTER TABLE pipeline_policy ADD COLUMN bootstrap_l3_worker_count INTEGER NOT NULL DEFAULT 1")
    if "bootstrap_l3_queue_max" not in policy_cols:
        conn.execute("ALTER TABLE pipeline_policy ADD COLUMN bootstrap_l3_queue_max INTEGER NOT NULL DEFAULT 1000")
    if "bootstrap_exit_min_l2_coverage_bps" not in policy_cols:
        conn.execute("ALTER TABLE pipeline_policy ADD COLUMN bootstrap_exit_min_l2_coverage_bps INTEGER NOT NULL DEFAULT 9500")
    if "bootstrap_exit_max_sec" not in policy_cols:
        conn.execute("ALTER TABLE pipeline_policy ADD COLUMN bootstrap_exit_max_sec INTEGER NOT NULL DEFAULT 1800")
    error_cols = _table_columns(conn, "pipeline_error_events")
    if "scope_type" not in error_cols:
        conn.execute(
            "ALTER TABLE pipeline_error_events ADD COLUMN scope_type TEXT NOT NULL DEFAULT 'GLOBAL' CHECK (scope_type IN ('GLOBAL', 'REPO'))"
        )
    conn.execute(
        """
        UPDATE pipeline_error_events
        SET scope_type = CASE
            WHEN repo_root IS NULL OR TRIM(repo_root) = '' THEN 'GLOBAL'
            ELSE 'REPO'
        END
        """
    )
    conn.execute(
        """
        CREATE INDEX IF NOT EXISTS idx_pipeline_error_events_scope_time
        ON pipeline_error_events(scope_type, occurred_at DESC)
        """
    )


def _fallback_upgrade_0003(conn: sqlite3.Connection) -> None:
    """0003 리비전 컬럼을 적용한다."""
    matrix_cols = _table_columns(conn, "pipeline_lsp_matrix_runs")
    if "required_languages_json" not in matrix_cols:
        conn.execute("ALTER TABLE pipeline_lsp_matrix_runs ADD COLUMN required_languages_json TEXT NOT NULL DEFAULT '[]'")
    if "fail_on_unavailable" not in matrix_cols:
        conn.execute("ALTER TABLE pipeline_lsp_matrix_runs ADD COLUMN fail_on_unavailable INTEGER NOT NULL DEFAULT 1")
    if "strict_symbol_gate" not in matrix_cols:
        conn.execute("ALTER TABLE pipeline_lsp_matrix_runs ADD COLUMN strict_symbol_gate INTEGER NOT NULL DEFAULT 1")
    if "started_at" not in matrix_cols:
        conn.execute("ALTER TABLE pipeline_lsp_matrix_runs ADD COLUMN started_at TEXT NOT NULL DEFAULT ''")
    if "finished_at" not in matrix_cols:
        conn.execute("ALTER TABLE pipeline_lsp_matrix_runs ADD COLUMN finished_at TEXT NULL")
    if "status" not in matrix_cols:
        conn.execute("ALTER TABLE pipeline_lsp_matrix_runs ADD COLUMN status TEXT NOT NULL DEFAULT 'RUNNING'")
    if "summary_json" not in matrix_cols:
        conn.execute("ALTER TABLE pipeline_lsp_matrix_runs ADD COLUMN summary_json TEXT NULL")

    probe_cols = _table_columns(conn, "language_probe_status")
    if "enabled" not in probe_cols:
        conn.execute("ALTER TABLE language_probe_status ADD COLUMN enabled INTEGER NOT NULL DEFAULT 1")
    if "available" not in probe_cols:
        conn.execute("ALTER TABLE language_probe_status ADD COLUMN available INTEGER NOT NULL DEFAULT 0")
    if "last_probe_at" not in probe_cols:
        conn.execute("ALTER TABLE language_probe_status ADD COLUMN last_probe_at TEXT NULL")
    if "last_error_code" not in probe_cols:
        conn.execute("ALTER TABLE language_probe_status ADD COLUMN last_error_code TEXT NULL")
    if "last_error_message" not in probe_cols:
        conn.execute("ALTER TABLE language_probe_status ADD COLUMN last_error_message TEXT NULL")
    if "updated_at" not in probe_cols:
        conn.execute("ALTER TABLE language_probe_status ADD COLUMN updated_at TEXT NOT NULL DEFAULT ''")

    symbol_cols = _table_columns(conn, "lsp_symbols")
    if "symbol_key" not in symbol_cols:
        conn.execute("ALTER TABLE lsp_symbols ADD COLUMN symbol_key TEXT NULL")
    if "parent_symbol_key" not in symbol_cols:
        conn.execute("ALTER TABLE lsp_symbols ADD COLUMN parent_symbol_key TEXT NULL")
    if "depth" not in symbol_cols:
        conn.execute("ALTER TABLE lsp_symbols ADD COLUMN depth INTEGER NOT NULL DEFAULT 0")
    if "container_name" not in symbol_cols:
        conn.execute("ALTER TABLE lsp_symbols ADD COLUMN container_name TEXT NULL")
    conn.execute(
        """
        CREATE INDEX IF NOT EXISTS idx_lsp_symbols_repo_path_depth
        ON lsp_symbols(repo_root, relative_path, depth, line)
        """
    )
    conn.execute(
        """
        CREATE INDEX IF NOT EXISTS idx_lsp_symbols_symbol_key
        ON lsp_symbols(repo_root, relative_path, content_hash, symbol_key)
        """
    )


def _fallback_upgrade_0004(conn: sqlite3.Connection) -> None:
    """0004 리비전 컬럼을 적용한다."""
    registry_cols = _table_columns(conn, "daemon_registry")
    if "deployment_state" not in registry_cols:
        conn.execute(
            "ALTER TABLE daemon_registry ADD COLUMN deployment_state TEXT NOT NULL DEFAULT 'ACTIVE'"
        )
    if "health_fail_streak" not in registry_cols:
        conn.execute(
            "ALTER TABLE daemon_registry ADD COLUMN health_fail_streak INTEGER NOT NULL DEFAULT 0"
        )
    if "last_health_error" not in registry_cols:
        conn.execute("ALTER TABLE daemon_registry ADD COLUMN last_health_error TEXT NULL")
    if "last_health_at" not in registry_cols:
        conn.execute("ALTER TABLE daemon_registry ADD COLUMN last_health_at TEXT NULL")
    conn.execute(
        """
        CREATE INDEX IF NOT EXISTS idx_daemon_registry_workspace
        ON daemon_registry(workspace_root, is_draining, deployment_state, last_seen_at DESC)
        """
    )



import logging
import os
import pathlib
import platform
import shutil
import subprocess

from overrides import override

from solidlsp.ls import SolidLanguageServer
from solidlsp.ls_config import LanguageServerConfig
from solidlsp.lsp_protocol_handler.lsp_types import InitializeParams
from solidlsp.lsp_protocol_handler.server import ProcessLaunchInfo
from solidlsp.settings import SolidLSPSettings

log = logging.getLogger(__name__)


class JuliaLanguageServer(SolidLanguageServer):
    """
    Language server implementation for Julia using LanguageServer.jl.
    """

    def __init__(self, config: LanguageServerConfig, repository_root_path: str, solidlsp_settings: SolidLSPSettings):
        julia_executable = self._setup_runtime_dependency()  # PASS LOGGER
        julia_code = "using LanguageServer; runserver()"

        julia_ls_cmd: str | list[str]
        if platform.system() == "Windows":
            # On Windows, pass as list (Serena handles shell=True differently)
            julia_ls_cmd = [julia_executable, "--startup-file=no", "--history-file=no", "-e", julia_code, repository_root_path]
        else:
            # On Linux/macOS, build shell-escaped string
            import shlex

            julia_ls_cmd = (
                f"{shlex.quote(julia_executable)} "
                f"--startup-file=no "
                f"--history-file=no "
                f"-e {shlex.quote(julia_code)} "
                f"{shlex.quote(repository_root_path)}"
            )

        log.info(f"[JULIA DEBUG] Command: {julia_ls_cmd}")

        super().__init__(
            config, repository_root_path, ProcessLaunchInfo(cmd=julia_ls_cmd, cwd=repository_root_path), "julia", solidlsp_settings
        )

    @staticmethod
    def _setup_runtime_dependency() -> str:
        """
        Check if the Julia runtime is available and return its full path.
        Raises RuntimeError with a helpful message if the dependency is missing.
        """
        # First check if julia is in PATH
        julia_path = shutil.which("julia")

        # If not found in PATH, check common installation locations
        if julia_path is None:
            common_locations = [
                os.path.expanduser("~/.juliaup/bin/julia"),
                os.path.expanduser("~/.julia/bin/julia"),
                "/usr/local/bin/julia",
                "/usr/bin/julia",
            ]

            for location in common_locations:
                if os.path.isfile(location) and os.access(location, os.X_OK):
                    julia_path = location
                    break

        if julia_path is None:
            raise RuntimeError(
                "Julia is not installed or not in your PATH. "
                "Please install Julia from https://julialang.org/downloads/ and ensure it is accessible. "
                f"Checked locations: {common_locations}"
            )

        # Check if LanguageServer.jl is installed
        check_cmd = [julia_path, "-e", "using LanguageServer"]
        try:
            result = subprocess.run(check_cmd, check=False, capture_output=True, text=True, timeout=10)
            if result.returncode != 0:
                # LanguageServer.jl not found, install it
                JuliaLanguageServer._install_language_server(julia_path)
        except subprocess.TimeoutExpired:
            # Assume it needs installation
            JuliaLanguageServer._install_language_server(julia_path)

        return julia_path

    @staticmethod
    def _install_language_server(julia_path: str) -> None:
        """Install LanguageServer.jl package."""
        log.info("LanguageServer.jl not found. Installing... (this may take a minute)")

        install_cmd = [julia_path, "-e", 'using Pkg; Pkg.add("LanguageServer")']

        try:
            result = subprocess.run(install_cmd, check=False, capture_output=True, text=True, timeout=300)  # 5 minutes for installation

            if result.returncode == 0:
                log.info("LanguageServer.jl installed successfully!")
            else:
                raise RuntimeError(f"Failed to install LanguageServer.jl: {result.stderr}")
        except subprocess.TimeoutExpired:
            raise RuntimeError(
                "LanguageServer.jl installation timed out. Please install manually: julia -e 'using Pkg; Pkg.add(\"LanguageServer\")'"
            )

    @override
    def is_ignored_dirname(self, dirname: str) -> bool:
        """Define language-specific directories to ignore for Julia projects."""
        return super().is_ignored_dirname(dirname) or dirname in [".julia", "build", "dist"]

    def _get_initialize_params(self, repository_absolute_path: str) -> InitializeParams:
        """
        Returns the initialize params for the Julia Language Server.
        """
        root_uri = pathlib.Path(repository_absolute_path).as_uri()
        initialize_params: InitializeParams = {  # type: ignore
            "processId": os.getpid(),
            "rootPath": repository_absolute_path,
            "rootUri": root_uri,
            "capabilities": {
                "workspace": {"workspaceFolders": True},
                "textDocument": {
                    "definition": {"dynamicRegistration": True},
                    "references": {"dynamicRegistration": True},
                    "documentSymbol": {"dynamicRegistration": True},
                },
            },
            "workspaceFolders": [
                {
                    "uri": root_uri,
                    "name": os.path.basename(repository_absolute_path),
                }
            ],
        }
        return initialize_params  # type: ignore

    def _start_server(self) -> None:
        """Start the LanguageServer.jl server process."""

        def do_nothing(params: dict) -> None:
            return

        def window_log_message(msg: dict) -> None:
            log.info(f"LSP: window/logMessage: {msg}")

        self.server.on_notification("window/logMessage", window_log_message)
        self.server.on_notification("$/progress", do_nothing)
        self.server.on_notification("textDocument/publishDiagnostics", do_nothing)

        log.info("Starting LanguageServer.jl server process")
        self.server.start()

        initialize_params = self._get_initialize_params(self.repository_root_path)
        log.info("Sending initialize request to Julia Language Server")

        init_response = self.server.send.initialize(initialize_params)
        assert "definitionProvider" in init_response["capabilities"]
        assert "referencesProvider" in init_response["capabilities"]
        assert "documentSymbolProvider" in init_response["capabilities"]

        self.server.notify.initialized({})
        log.info("Julia Language Server is initialized and ready.")



"""Vue language server integration."""

import logging
import os
import pathlib
import shutil
import threading
from pathlib import Path
from time import sleep

from overrides import override

from solidlsp import ls_types
from solidlsp.language_servers._adapter_common import ensure_commands_available
from solidlsp.language_servers.common import RuntimeDependency, RuntimeDependencyCollection
from solidlsp.language_servers.typescript_language_server import (
    TypeScriptLanguageServer,
    prefer_non_node_modules_definition,
)
from solidlsp.ls import LSPFileBuffer, SolidLanguageServer
from solidlsp.ls_config import Language, LanguageServerConfig
from solidlsp.ls_exceptions import SolidLSPException
from solidlsp.ls_types import Location
from solidlsp.ls_utils import PathUtils
from solidlsp.lsp_protocol_handler import lsp_types
from solidlsp.lsp_protocol_handler.lsp_types import DocumentSymbol, ExecuteCommandParams, InitializeParams, SymbolInformation
from solidlsp.lsp_protocol_handler.server import ProcessLaunchInfo
from solidlsp.settings import SolidLSPSettings

log = logging.getLogger(__name__)

class VueTypeScriptServer(TypeScriptLanguageServer):
    """TypeScript LS configured with @vue/typescript-plugin for Vue file support."""

    @classmethod
    @override
    def get_language_enum_instance(cls) -> Language:
        """Return TYPESCRIPT since this is a TypeScript language server variant.

        Note: VueTypeScriptServer is a companion server that uses TypeScript's language server
        with the Vue TypeScript plugin. It reports as TYPESCRIPT to maintain compatibility
        with the TypeScript language server infrastructure.
        """
        return Language.TYPESCRIPT

    class DependencyProvider(TypeScriptLanguageServer.DependencyProvider):
        override_ts_ls_executable: str | None = None

        def _get_or_install_core_dependency(self) -> str:
            if self.override_ts_ls_executable is not None:
                return self.override_ts_ls_executable
            return super()._get_or_install_core_dependency()

    @override
    def _get_language_id_for_file(self, relative_file_path: str) -> str:
        """Return the correct language ID for files.

        Vue files must be opened with language ID "vue" for the @vue/typescript-plugin
        to process them correctly. The plugin is configured with "languages": ["vue"]
        in the initialization options.
        """
        ext = os.path.splitext(relative_file_path)[1].lower()
        if ext == ".vue":
            return "vue"
        elif ext in (".ts", ".tsx", ".mts", ".cts"):
            return "typescript"
        elif ext in (".js", ".jsx", ".mjs", ".cjs"):
            return "javascript"
        else:
            return "typescript"

    def __init__(
        self,
        config: LanguageServerConfig,
        repository_root_path: str,
        solidlsp_settings: SolidLSPSettings,
        vue_plugin_path: str,
        tsdk_path: str,
        ts_ls_executable_path: str,
    ):
        self._vue_plugin_path = vue_plugin_path
        self._custom_tsdk_path = tsdk_path
        VueTypeScriptServer.DependencyProvider.override_ts_ls_executable = ts_ls_executable_path
        super().__init__(config, repository_root_path, solidlsp_settings)
        VueTypeScriptServer.DependencyProvider.override_ts_ls_executable = None

    @override
    def _get_initialize_params(self, repository_absolute_path: str) -> InitializeParams:
        params = super()._get_initialize_params(repository_absolute_path)

        params["initializationOptions"] = {
            "plugins": [
                {
                    "name": "@vue/typescript-plugin",
                    "location": self._vue_plugin_path,
                    "languages": ["vue"],
                }
            ],
            "tsserver": {
                "path": self._custom_tsdk_path,
            },
        }

        if "workspace" in params["capabilities"]:
            params["capabilities"]["workspace"]["executeCommand"] = {"dynamicRegistration": True}

        return params

    @override
    def _start_server(self) -> None:
        def workspace_configuration_handler(params: dict) -> list:
            items = params.get("items", [])
            return [{} for _ in items]

        self.server.on_request("workspace/configuration", workspace_configuration_handler)
        super()._start_server()

class VueLanguageServer(SolidLanguageServer):
    """
    Language server for Vue Single File Components using @vue/language-server (Volar) with companion TypeScript LS.

    You can pass the following entries in ls_specific_settings["vue"]:
        - vue_language_server_version: Version of @vue/language-server to install (default: "3.1.5")

    Note: TypeScript versions are configured via ls_specific_settings["typescript"]:
        - typescript_version: Version of TypeScript to install (default: "5.9.3")
        - typescript_language_server_version: Version of typescript-language-server to install (default: "5.1.3")
    """

    TS_SERVER_READY_TIMEOUT = 5.0
    VUE_SERVER_READY_TIMEOUT = 3.0
    VUE_INDEXING_WAIT_TIME = 4.0 if os.name == "nt" else 2.0

    def __init__(self, config: LanguageServerConfig, repository_root_path: str, solidlsp_settings: SolidLSPSettings):
        vue_lsp_executable_path, self.tsdk_path, self._ts_ls_cmd = self._setup_runtime_dependencies(config, solidlsp_settings)
        self._vue_ls_dir = os.path.join(self.ls_resources_dir(solidlsp_settings), "vue-lsp")
        super().__init__(
            config,
            repository_root_path,
            ProcessLaunchInfo(cmd=vue_lsp_executable_path, cwd=repository_root_path),
            "vue",
            solidlsp_settings,
        )
        self.server_ready = threading.Event()
        self.initialize_searcher_command_available = threading.Event()
        self._ts_server: VueTypeScriptServer | None = None
        self._ts_server_started = False
        self._vue_files_indexed = False
        self._indexed_vue_file_uris: list[str] = []

    @override
    def is_ignored_dirname(self, dirname: str) -> bool:
        return super().is_ignored_dirname(dirname) or dirname in [
            "node_modules",
            "dist",
            "build",
            "coverage",
            ".nuxt",
            ".output",
        ]

    @override
    def _get_language_id_for_file(self, relative_file_path: str) -> str:
        ext = os.path.splitext(relative_file_path)[1].lower()
        if ext == ".vue":
            return "vue"
        elif ext in (".ts", ".tsx", ".mts", ".cts"):
            return "typescript"
        elif ext in (".js", ".jsx", ".mjs", ".cjs"):
            return "javascript"
        else:
            return "vue"

    def _is_typescript_file(self, file_path: str) -> bool:
        ext = os.path.splitext(file_path)[1].lower()
        return ext in (".ts", ".tsx", ".mts", ".cts", ".js", ".jsx", ".mjs", ".cjs")

    def _find_all_vue_files(self) -> list[str]:
        vue_files = []
        repo_path = Path(self.repository_root_path)

        for vue_file in repo_path.rglob("*.vue"):
            try:
                relative_path = str(vue_file.relative_to(repo_path))
                if "node_modules" not in relative_path and not relative_path.startswith("."):
                    vue_files.append(relative_path)
            except (OSError, RuntimeError, ValueError, TypeError, SolidLSPException) as e:
                log.debug(f"Error processing Vue file {vue_file}: {e}")

        return vue_files

    def _ensure_vue_files_indexed_on_ts_server(self) -> None:
        if self._vue_files_indexed:
            return

        assert self._ts_server is not None
        log.info("Indexing .vue files on TypeScript server for cross-file references")
        vue_files = self._find_all_vue_files()
        log.debug(f"Found {len(vue_files)} .vue files to index")

        for vue_file in vue_files:
            try:
                with self._ts_server.open_file(vue_file) as file_buffer:
                    file_buffer.ref_count += 1
                    self._indexed_vue_file_uris.append(file_buffer.uri)
            except (OSError, RuntimeError, ValueError, TypeError, SolidLSPException) as e:
                log.debug(f"Failed to open {vue_file} on TS server: {e}")

        self._vue_files_indexed = True
        log.info("Vue file indexing on TypeScript server complete")

        sleep(self._get_vue_indexing_wait_time())
        log.debug("Wait period after Vue file indexing complete")

    def _get_vue_indexing_wait_time(self) -> float:
        return self.VUE_INDEXING_WAIT_TIME

    def _send_references_request(self, relative_file_path: str, line: int, column: int) -> list[lsp_types.Location] | None:
        uri = PathUtils.path_to_uri(os.path.join(self.repository_root_path, relative_file_path))
        request_params = {
            "textDocument": {"uri": uri},
            "position": {"line": line, "character": column},
            "context": {"includeDeclaration": False},
        }

        return self.server.send.references(request_params)  # type: ignore[arg-type]

    def _send_ts_references_request(self, relative_file_path: str, line: int, column: int) -> list[ls_types.Location]:
        assert self._ts_server is not None
        uri = PathUtils.path_to_uri(os.path.join(self.repository_root_path, relative_file_path))
        request_params = {
            "textDocument": {"uri": uri},
            "position": {"line": line, "character": column},
            "context": {"includeDeclaration": True},
        }

        with self._ts_server.open_file(relative_file_path):
            response = self._ts_server.handler.send.references(request_params)  # type: ignore[arg-type]

        result: list[ls_types.Location] = []
        if response is not None:
            for item in response:
                abs_path = PathUtils.uri_to_path(item["uri"])
                if not Path(abs_path).is_relative_to(self.repository_root_path):
                    log.debug(f"Found reference outside repository: {abs_path}, skipping")
                    continue

                rel_path = Path(abs_path).relative_to(self.repository_root_path)
                if self.is_ignored_path(str(rel_path)):
                    log.debug(f"Ignoring reference in {rel_path}")
                    continue

                new_item: dict = {}
                new_item.update(item)  # type: ignore[arg-type]
                new_item["absolutePath"] = str(abs_path)
                new_item["relativePath"] = str(rel_path)
                result.append(ls_types.Location(**new_item))  # type: ignore

        return result

    def request_file_references(self, relative_file_path: str) -> list:
        if not self.server_started:
            log.error("request_file_references called before Language Server started")
            raise SolidLSPException("Language Server not started")

        absolute_file_path = os.path.join(self.repository_root_path, relative_file_path)
        uri = PathUtils.path_to_uri(absolute_file_path)

        request_params = {"textDocument": {"uri": uri}}

        log.info(f"Sending volar/client/findFileReference request for {relative_file_path}")
        log.info(f"Request URI: {uri}")
        log.info(f"Request params: {request_params}")

        try:
            with self.open_file(relative_file_path):
                log.debug(f"Sending volar/client/findFileReference for {relative_file_path}")
                log.debug(f"Request params: {request_params}")

                response = self.server.send_request("volar/client/findFileReference", request_params)

                log.debug(f"Received response type: {type(response)}")

            log.info(f"Received file references response: {response}")
            log.info(f"Response type: {type(response)}")

            if response is None:
                log.debug(f"No file references found for {relative_file_path}")
                return []

            if not isinstance(response, list):
                log.warning(f"Unexpected response format from volar/client/findFileReference: {type(response)}")
                return []

            ret: list[Location] = []
            for item in response:
                if not isinstance(item, dict) or "uri" not in item:
                    log.debug(f"Skipping invalid location item: {item}")
                    continue

                abs_path = PathUtils.uri_to_path(item["uri"])  # type: ignore[arg-type]
                if not Path(abs_path).is_relative_to(self.repository_root_path):
                    log.warning(f"Found file reference outside repository: {abs_path}, skipping")
                    continue

                rel_path = Path(abs_path).relative_to(self.repository_root_path)
                if self.is_ignored_path(str(rel_path)):
                    log.debug(f"Ignoring file reference in {rel_path}")
                    continue

                new_item: dict = {}
                new_item.update(item)  # type: ignore[arg-type]
                new_item["absolutePath"] = str(abs_path)
                new_item["relativePath"] = str(rel_path)
                ret.append(Location(**new_item))  # type: ignore

            log.debug(f"Found {len(ret)} file references for {relative_file_path}")
            return ret

        except (OSError, RuntimeError, ValueError, TypeError, SolidLSPException) as e:
            log.warning(f"Error requesting file references for {relative_file_path}: {e}")
            return []

    @override
    def request_references(self, relative_file_path: str, line: int, column: int) -> list[ls_types.Location]:
        if not self.server_started:
            log.error("request_references called before Language Server started")
            raise SolidLSPException("Language Server not started")

        if not self._has_waited_for_cross_file_references:
            sleep(self._get_wait_time_for_cross_file_referencing())
            self._has_waited_for_cross_file_references = True

        self._ensure_vue_files_indexed_on_ts_server()
        symbol_refs = self._send_ts_references_request(relative_file_path, line=line, column=column)

        if relative_file_path.endswith(".vue"):
            log.info(f"Attempting to find file-level references for Vue component {relative_file_path}")
            file_refs = self.request_file_references(relative_file_path)
            log.info(f"file_refs result: {len(file_refs)} references found")

            seen = set()
            for ref in symbol_refs:
                key = (ref["uri"], ref["range"]["start"]["line"], ref["range"]["start"]["character"])
                seen.add(key)

            for file_ref in file_refs:
                key = (file_ref["uri"], file_ref["range"]["start"]["line"], file_ref["range"]["start"]["character"])
                if key not in seen:
                    symbol_refs.append(file_ref)
                    seen.add(key)

            log.info(f"Total references for {relative_file_path}: {len(symbol_refs)} (symbol refs + file refs, deduplicated)")

        return symbol_refs

    @override
    def request_definition(self, relative_file_path: str, line: int, column: int) -> list[ls_types.Location]:
        if not self.server_started:
            log.error("request_definition called before Language Server started")
            raise SolidLSPException("Language Server not started")

        assert self._ts_server is not None
        with self._ts_server.open_file(relative_file_path):
            return self._ts_server.request_definition(relative_file_path, line, column)

    @override
    def request_rename_symbol_edit(self, relative_file_path: str, line: int, column: int, new_name: str) -> ls_types.WorkspaceEdit | None:
        if not self.server_started:
            log.error("request_rename_symbol_edit called before Language Server started")
            raise SolidLSPException("Language Server not started")

        assert self._ts_server is not None
        with self._ts_server.open_file(relative_file_path):
            return self._ts_server.request_rename_symbol_edit(relative_file_path, line, column, new_name)

    @classmethod
    def _setup_runtime_dependencies(cls, config: LanguageServerConfig, solidlsp_settings: SolidLSPSettings) -> tuple[list[str], str, str]:
        ensure_commands_available(["node", "npm"])

        typescript_config = solidlsp_settings.get_ls_specific_settings(Language.TYPESCRIPT)
        typescript_version = typescript_config.get("typescript_version", "5.9.3")
        typescript_language_server_version = typescript_config.get("typescript_language_server_version", "5.1.3")
        vue_config = solidlsp_settings.get_ls_specific_settings(Language.VUE)
        vue_language_server_version = vue_config.get("vue_language_server_version", "3.1.5")

        deps = RuntimeDependencyCollection(
            [
                RuntimeDependency(
                    id="vue-language-server",
                    description="Vue language server package (Volar)",
                    command=["npm", "install", "--prefix", "./", f"@vue/language-server@{vue_language_server_version}"],
                    platform_id="any",
                ),
                RuntimeDependency(
                    id="typescript",
                    description="TypeScript (required for tsdk)",
                    command=["npm", "install", "--prefix", "./", f"typescript@{typescript_version}"],
                    platform_id="any",
                ),
                RuntimeDependency(
                    id="typescript-language-server",
                    description="TypeScript language server (for Vue LS 3.x tsserver forwarding)",
                    command=[
                        "npm",
                        "install",
                        "--prefix",
                        "./",
                        f"typescript-language-server@{typescript_language_server_version}",
                    ],
                    platform_id="any",
                ),
            ]
        )

        vue_ls_dir = os.path.join(cls.ls_resources_dir(solidlsp_settings), "vue-lsp")
        vue_executable_path = os.path.join(vue_ls_dir, "node_modules", ".bin", "vue-language-server")
        ts_ls_executable_path = os.path.join(vue_ls_dir, "node_modules", ".bin", "typescript-language-server")

        if os.name == "nt":
            vue_executable_path += ".cmd"
            ts_ls_executable_path += ".cmd"

        tsdk_path = os.path.join(vue_ls_dir, "node_modules", "typescript", "lib")

        version_file = os.path.join(vue_ls_dir, ".installed_version")
        expected_version = f"{vue_language_server_version}_{typescript_version}_{typescript_language_server_version}"

        needs_install = False
        if not os.path.exists(vue_executable_path) or not os.path.exists(ts_ls_executable_path):
            log.info("Vue/TypeScript Language Server executables not found.")
            needs_install = True
        elif os.path.exists(version_file):
            with open(version_file) as f:
                installed_version = f.read().strip()
            if installed_version != expected_version:
                log.info(
                    f"Vue Language Server version mismatch: installed={installed_version}, expected={expected_version}. Reinstalling..."
                )
                needs_install = True
        else:
            log.info("Vue Language Server version file not found. Reinstalling to ensure correct version...")
            needs_install = True

        if needs_install:
            log.info("Installing Vue/TypeScript Language Server dependencies...")
            deps.install(vue_ls_dir)
            with open(version_file, "w") as f:
                f.write(expected_version)
            log.info("Vue language server dependencies installed successfully")

        if not os.path.exists(vue_executable_path):
            raise FileNotFoundError(
                f"vue-language-server executable not found at {vue_executable_path}, something went wrong with the installation."
            )

        if not os.path.exists(ts_ls_executable_path):
            raise FileNotFoundError(
                f"typescript-language-server executable not found at {ts_ls_executable_path}, something went wrong with the installation."
            )

        return [vue_executable_path, "--stdio"], tsdk_path, ts_ls_executable_path

    def _get_initialize_params(self, repository_absolute_path: str) -> InitializeParams:
        root_uri = pathlib.Path(repository_absolute_path).as_uri()
        initialize_params = {
            "locale": "en",
            "capabilities": {
                "textDocument": {
                    "synchronization": {"didSave": True, "dynamicRegistration": True},
                    "completion": {"dynamicRegistration": True, "completionItem": {"snippetSupport": True}},
                    "definition": {"dynamicRegistration": True, "linkSupport": True},
                    "references": {"dynamicRegistration": True},
                    "documentSymbol": {
                        "dynamicRegistration": True,
                        "hierarchicalDocumentSymbolSupport": True,
                        "symbolKind": {"valueSet": list(range(1, 27))},
                    },
                    "hover": {"dynamicRegistration": True, "contentFormat": ["markdown", "plaintext"]},
                    "signatureHelp": {"dynamicRegistration": True},
                    "codeAction": {"dynamicRegistration": True},
                    "rename": {"dynamicRegistration": True, "prepareSupport": True},
                },
                "workspace": {
                    "workspaceFolders": True,
                    "didChangeConfiguration": {"dynamicRegistration": True},
                    "symbol": {"dynamicRegistration": True},
                },
            },
            "processId": os.getpid(),
            "rootPath": repository_absolute_path,
            "rootUri": root_uri,
            "workspaceFolders": [
                {
                    "uri": root_uri,
                    "name": os.path.basename(repository_absolute_path),
                }
            ],
            "initializationOptions": {
                "vue": {
                    "hybridMode": True,
                },
                "typescript": {
                    "tsdk": self.tsdk_path,
                },
            },
        }
        return initialize_params  # type: ignore

    def _start_typescript_server(self) -> None:
        try:
            vue_ts_plugin_path = os.path.join(self._vue_ls_dir, "node_modules", "@vue", "typescript-plugin")

            ts_config = LanguageServerConfig(
                code_language=Language.TYPESCRIPT,
                trace_lsp_communication=False,
            )

            log.info("Creating companion VueTypeScriptServer")
            self._ts_server = VueTypeScriptServer(
                config=ts_config,
                repository_root_path=self.repository_root_path,
                solidlsp_settings=self._solidlsp_settings,
                vue_plugin_path=vue_ts_plugin_path,
                tsdk_path=self.tsdk_path,
                ts_ls_executable_path=self._ts_ls_cmd,
            )

            log.info("Starting companion TypeScript server")
            self._ts_server.start()

            log.info("Waiting for companion TypeScript server to be ready...")
            if not self._ts_server.server_ready.wait(timeout=self.TS_SERVER_READY_TIMEOUT):
                log.warning(
                    f"Timeout waiting for companion TypeScript server to be ready after {self.TS_SERVER_READY_TIMEOUT} seconds, proceeding anyway"
                )
                self._ts_server.server_ready.set()

            self._ts_server_started = True
            log.info("Companion TypeScript server ready")
        except (OSError, RuntimeError, ValueError, TypeError, SolidLSPException) as e:
            log.error(f"Error starting TypeScript server: {e}")
            self._ts_server = None
            self._ts_server_started = False
            raise

    def _forward_tsserver_request(self, method: str, params: dict[str, object]) -> object | None:
        if self._ts_server is None:
            log.error("Cannot forward tsserver request - TypeScript server not started")
            return None

        try:
            execute_params: ExecuteCommandParams = {
                "command": "typescript.tsserverRequest",
                "arguments": [method, params, {"isAsync": True, "lowPriority": True}],
            }
            result = self._ts_server.handler.send.execute_command(execute_params)
            log.debug(f"TypeScript server raw response for {method}: {result}")

            if isinstance(result, dict) and "body" in result:
                return result["body"]
            return result
        except (OSError, RuntimeError, ValueError, TypeError, SolidLSPException) as e:
            log.error(f"Error forwarding tsserver request {method}: {e}")
            return None

    def _cleanup_indexed_vue_files(self) -> None:
        if not self._indexed_vue_file_uris or self._ts_server is None:
            return

        log.debug(f"Cleaning up {len(self._indexed_vue_file_uris)} indexed Vue files")
        for uri in self._indexed_vue_file_uris:
            try:
                if uri in self._ts_server.open_file_buffers:
                    file_buffer = self._ts_server.open_file_buffers[uri]
                    file_buffer.ref_count -= 1

                    if file_buffer.ref_count == 0:
                        self._ts_server.server.notify.did_close_text_document({"textDocument": {"uri": uri}})
                        del self._ts_server.open_file_buffers[uri]
                        log.debug(f"Closed indexed Vue file: {uri}")
            except (OSError, RuntimeError, ValueError, TypeError, SolidLSPException) as e:
                log.debug(f"Error closing indexed Vue file {uri}: {e}")

        self._indexed_vue_file_uris.clear()

    def _stop_typescript_server(self) -> None:
        if self._ts_server is not None:
            try:
                log.info("Stopping companion TypeScript server")
                self._ts_server.stop()
            except (OSError, RuntimeError, ValueError, TypeError, SolidLSPException) as e:
                log.warning(f"Error stopping TypeScript server: {e}")
            finally:
                self._ts_server = None
                self._ts_server_started = False

    @override
    def _start_server(self) -> None:
        self._start_typescript_server()

        def register_capability_handler(params: dict) -> None:
            assert "registrations" in params
            for registration in params["registrations"]:
                if registration["method"] == "workspace/executeCommand":
                    self.initialize_searcher_command_available.set()
            return

        def configuration_handler(params: dict) -> list:
            items = params.get("items", [])
            return [{} for _ in items]

        def do_nothing(params: dict) -> None:
            return

        def window_log_message(msg: dict) -> None:
            log.info(f"LSP: window/logMessage: {msg}")
            message_text = msg.get("message", "")
            if "initialized" in message_text.lower() or "ready" in message_text.lower():
                log.info("Vue language server ready signal detected")
                self.server_ready.set()

        def tsserver_request_notification_handler(params: list) -> None:
            try:
                if params and len(params) > 0 and len(params[0]) >= 2:
                    request_id = params[0][0]
                    method = params[0][1]
                    method_params = params[0][2] if len(params[0]) > 2 else {}
                    log.debug(f"Received tsserver/request: id={request_id}, method={method}")

                    if method == "_vue:projectInfo":
                        file_path = method_params.get("file", "")
                        tsconfig_path = self._find_tsconfig_for_file(file_path)
                        result = {"configFileName": tsconfig_path} if tsconfig_path else None
                        response = [[request_id, result]]
                        self.server.notify.send_notification("tsserver/response", response)
                        log.debug(f"Sent tsserver/response for projectInfo: {tsconfig_path}")
                    else:
                        result = self._forward_tsserver_request(method, method_params)
                        response = [[request_id, result]]
                        self.server.notify.send_notification("tsserver/response", response)
                        log.debug(f"Forwarded tsserver/response for {method}: {result}")
                else:
                    log.warning(f"Unexpected tsserver/request params format: {params}")
            except (OSError, RuntimeError, ValueError, TypeError, SolidLSPException) as e:
                log.error(f"Error handling tsserver/request: {e}")

        self.server.on_request("client/registerCapability", register_capability_handler)
        self.server.on_request("workspace/configuration", configuration_handler)
        self.server.on_notification("tsserver/request", tsserver_request_notification_handler)
        self.server.on_notification("window/logMessage", window_log_message)
        self.server.on_notification("$/progress", do_nothing)
        self.server.on_notification("textDocument/publishDiagnostics", do_nothing)

        log.info("Starting Vue server process")
        self.server.start()
        initialize_params = self._get_initialize_params(self.repository_root_path)

        log.info("Sending initialize request from LSP client to LSP server and awaiting response")
        init_response = self.server.send.initialize(initialize_params)
        log.debug(f"Received initialize response from Vue server: {init_response}")

        assert init_response["capabilities"]["textDocumentSync"] in [1, 2]

        self.server.notify.initialized({})

        log.info("Waiting for Vue language server to be ready...")
        if not self.server_ready.wait(timeout=self.VUE_SERVER_READY_TIMEOUT):
            log.info("Timeout waiting for Vue server ready signal, proceeding anyway")
            self.server_ready.set()
        else:
            log.info("Vue server initialization complete")

    def _find_tsconfig_for_file(self, file_path: str) -> str | None:
        if not file_path:
            tsconfig_path = os.path.join(self.repository_root_path, "tsconfig.json")
            return tsconfig_path if os.path.exists(tsconfig_path) else None

        current_dir = os.path.dirname(file_path)
        repo_root = os.path.abspath(self.repository_root_path)

        while current_dir and current_dir.startswith(repo_root):
            tsconfig_path = os.path.join(current_dir, "tsconfig.json")
            if os.path.exists(tsconfig_path):
                return tsconfig_path
            parent = os.path.dirname(current_dir)
            if parent == current_dir:
                break
            current_dir = parent

        tsconfig_path = os.path.join(repo_root, "tsconfig.json")
        return tsconfig_path if os.path.exists(tsconfig_path) else None

    @override
    def _get_wait_time_for_cross_file_referencing(self) -> float:
        return 5.0

    @override
    def stop(self, shutdown_timeout: float = 5.0) -> None:
        self._cleanup_indexed_vue_files()
        self._stop_typescript_server()
        super().stop(shutdown_timeout)

    @override
    def _get_preferred_definition(self, definitions: list[ls_types.Location]) -> ls_types.Location:
        return prefer_non_node_modules_definition(definitions)

    @override
    def _request_document_symbols(
        self, relative_file_path: str, file_data: LSPFileBuffer | None
    ) -> list[SymbolInformation] | list[DocumentSymbol] | None:
        """
        Override to filter out shorthand property references in Vue files.

        In Vue, when using shorthand syntax in defineExpose like `defineExpose({ pressCount })`,
        the Vue LSP returns both:
        - The Variable definition (e.g., `const pressCount = ref(0)`)
        - A Property symbol for the shorthand reference (e.g., `pressCount` in defineExpose)

        This causes duplicate symbols with the same name, which breaks symbol lookup.
        We filter out Property symbols that have a matching Variable with the same name
        at a different location (the definition), keeping only the definition.
        """
        symbols = super()._request_document_symbols(relative_file_path, file_data)

        if symbols is None or len(symbols) == 0:
            return symbols

        if not isinstance(symbols[0], dict) or "range" not in symbols[0]:
            return symbols

        return self._filter_shorthand_property_duplicates(symbols)

    def _filter_shorthand_property_duplicates(
        self, symbols: list[DocumentSymbol] | list[SymbolInformation]
    ) -> list[DocumentSymbol] | list[SymbolInformation]:
        """
        Filter out Property symbols that have a matching Variable symbol with the same name.

        This handles Vue's shorthand property syntax in defineExpose, where the same
        identifier appears as both a Variable definition and a Property reference.
        """
        VARIABLE_KIND = 13  # SymbolKind.Variable
        PROPERTY_KIND = 7  # SymbolKind.Property

        def filter_symbols(syms: list[dict]) -> list[dict]:
            variable_names: dict[str, set[int]] = {}
            for sym in syms:
                if sym.get("kind") == VARIABLE_KIND:
                    name = sym.get("name", "")
                    line = sym.get("range", {}).get("start", {}).get("line", -1)
                    if name not in variable_names:
                        variable_names[name] = set()
                    variable_names[name].add(line)

            filtered = []
            for sym in syms:
                name = sym.get("name", "")
                kind = sym.get("kind")
                line = sym.get("range", {}).get("start", {}).get("line", -1)

                if kind == PROPERTY_KIND and name in variable_names:
                    var_lines = variable_names[name]
                    if any(var_line != line for var_line in var_lines):
                        log.debug(
                            f"Filtering shorthand property reference '{name}' at line {line} "
                            f"(Variable definition exists at line(s) {var_lines})"
                        )
                        continue

                children = sym.get("children", [])
                if children:
                    sym = dict(sym)  # Create a copy to avoid mutating the original
                    sym["children"] = filter_symbols(children)

                filtered.append(sym)

            return filtered

        return filter_symbols(list(symbols))  # type: ignore



from __future__ import annotations
import json
from dataclasses import dataclass
from datetime import datetime, timezone
@dataclass(frozen=True)
class WorkspaceDTO:
    path: str
    name: str | None
    indexed_at: str | None
    is_active: bool
    def to_sql_params(self) -> dict[str, object]:
        return {
            "path": self.path,
            "name": self.name,
            "indexed_at": self.indexed_at,
            "is_active": 1 if self.is_active else 0,
        }
Workspace = WorkspaceDTO
@dataclass(frozen=True)
class DaemonRuntimeDTO:
    pid: int
    host: str
    port: int
    state: str
    started_at: str
    session_count: int
    last_heartbeat_at: str
    last_exit_reason: str | None = None
    def to_sql_params(self) -> dict[str, object]:
        return {
            "singleton_key": "default",
            "pid": self.pid,
            "host": self.host,
            "port": self.port,
            "state": self.state,
            "started_at": self.started_at,
            "session_count": self.session_count,
            "last_heartbeat_at": self.last_heartbeat_at,
            "last_exit_reason": self.last_exit_reason,
        }
@dataclass(frozen=True)
class DaemonRegistryEntryDTO:
    daemon_id: str
    host: str
    port: int
    pid: int
    workspace_root: str
    protocol: str
    started_at: str
    last_seen_at: str
    is_draining: bool
    deployment_state: str = "ACTIVE"
    health_fail_streak: int = 0
    last_health_error: str | None = None
    last_health_at: str | None = None
    def to_sql_params(self) -> dict[str, object]:
        return {
            "daemon_id": self.daemon_id,
            "host": self.host,
            "port": self.port,
            "pid": self.pid,
            "workspace_root": self.workspace_root,
            "protocol": self.protocol,
            "started_at": self.started_at,
            "last_seen_at": self.last_seen_at,
            "is_draining": 1 if self.is_draining else 0,
            "deployment_state": self.deployment_state,
            "health_fail_streak": self.health_fail_streak,
            "last_health_error": self.last_health_error,
            "last_health_at": self.last_health_at,
        }
@dataclass(frozen=True)
class HealthResponseDTO:
    status: str
    version: str
    uptime_sec: float
@dataclass(frozen=True)
class ErrorResponseDTO:
    code: str
    message: str
@dataclass(frozen=True)
class LanguageProbeStatusDTO:
    language: str
    enabled: bool
    available: bool
    last_probe_at: str | None
    last_error_code: str | None
    last_error_message: str | None
    updated_at: str
    symbol_extract_success: bool = False
    document_symbol_count: int = 0
    path_mapping_ok: bool = False
    timeout_occurred: bool = False
    recovered_by_restart: bool = False
    provisioning_mode: str | None = None
    missing_dependency: str | None = None
    install_hint: str | None = None
    def to_sql_params(self) -> dict[str, object]:
        return {
            "language": self.language,
            "enabled": 1 if self.enabled else 0,
            "available": 1 if self.available else 0,
            "last_probe_at": self.last_probe_at,
            "last_error_code": self.last_error_code,
            "last_error_message": self.last_error_message,
            "symbol_extract_success": 1 if self.symbol_extract_success else 0,
            "document_symbol_count": self.document_symbol_count,
            "path_mapping_ok": 1 if self.path_mapping_ok else 0,
            "timeout_occurred": 1 if self.timeout_occurred else 0,
            "recovered_by_restart": 1 if self.recovered_by_restart else 0,
            "updated_at": self.updated_at,
        }
    def to_dict(self) -> dict[str, object]:
        return {
            "language": self.language,
            "enabled": self.enabled,
            "available": self.available,
            "last_probe_at": self.last_probe_at,
            "last_error_code": self.last_error_code,
            "last_error_message": self.last_error_message,
            "symbol_extract_success": self.symbol_extract_success,
            "document_symbol_count": self.document_symbol_count,
            "path_mapping_ok": self.path_mapping_ok,
            "timeout_occurred": self.timeout_occurred,
            "recovered_by_restart": self.recovered_by_restart,
            "provisioning_mode": self.provisioning_mode,
            "missing_dependency": self.missing_dependency,
            "install_hint": self.install_hint,
        }
@dataclass(frozen=True)
class SearchErrorDTO:
    code: str
    message: str
    severity: str
    origin: str
    def to_dict(self) -> dict[str, str]:
        return {
            "code": self.code,
            "message": self.message,
            "severity": self.severity,
            "origin": self.origin,
        }
@dataclass(frozen=True)
class SearchPlaceholderResponseDTO:
    phase: str
    message: str
    query: str
    limit: int
@dataclass(frozen=True)
class CandidateFileDTO:
    repo_root: str
    relative_path: str
    score: float
    file_hash: str
@dataclass(frozen=True)
class CandidateIndexChangeDTO:
    repo_root: str
    relative_path: str
    absolute_path: str
    content_hash: str
    mtime_ns: int
    size_bytes: int
    event_source: str
    recorded_at: str
    def to_sql_params(self) -> dict[str, object]:
        return {
            "repo_root": self.repo_root,
            "relative_path": self.relative_path,
            "absolute_path": self.absolute_path,
            "content_hash": self.content_hash,
            "mtime_ns": self.mtime_ns,
            "size_bytes": self.size_bytes,
            "event_source": self.event_source,
            "recorded_at": self.recorded_at,
        }
@dataclass(frozen=True)
class CandidateIndexChangeLogDTO:
    change_id: int
    change_type: str
    status: str
    repo_root: str
    relative_path: str
    absolute_path: str | None
    content_hash: str | None
    mtime_ns: int | None
    size_bytes: int | None
    event_source: str
    reason: str | None
    created_at: str
    updated_at: str
@dataclass(frozen=True)
class RankingComponentsDTO:
    """검색 점수 구성요소를 표현한다."""

    rrf: float = 0.0
    importance: float = 0.0
    vector: float = 0.0
    hierarchy: float = 0.0
    final: float = 0.0

    def to_dict(self) -> dict[str, float]:
        """직렬화 가능한 점수 구성요소 딕셔너리를 반환한다."""
        return {
            "rrf": self.rrf,
            "importance": self.importance,
            "vector": self.vector,
            "hierarchy": self.hierarchy,
            "final": self.final,
        }


@dataclass(frozen=True)
class SearchItemDTO:
    item_type: str
    repo: str
    relative_path: str
    score: float
    source: str
    name: str | None
    kind: str | None
    content_hash: str | None = None
    rrf_score: float = 0.0
    importance_score: float = 0.0
    base_rrf_score: float = 0.0
    importance_norm_score: float = 0.0
    vector_norm_score: float = 0.0
    hierarchy_score: float = 0.0
    hierarchy_norm_score: float = 0.0
    symbol_key: str | None = None
    parent_symbol_key: str | None = None
    depth: int = 0
    container_name: str | None = None
    ranking_components: RankingComponentsDTO | None = None
    vector_score: float | None = None
    blended_score: float = 0.0
    final_score: float = 0.0
@dataclass(frozen=True)
class CollectedFileL1DTO:
    repo_root: str
    relative_path: str
    absolute_path: str
    repo_label: str
    mtime_ns: int
    size_bytes: int
    content_hash: str
    is_deleted: bool
    last_seen_at: str
    updated_at: str
    enrich_state: str
    def to_sql_params(self) -> dict[str, object]:
        return {
            "repo_root": self.repo_root,
            "relative_path": self.relative_path,
            "absolute_path": self.absolute_path,
            "repo_label": self.repo_label,
            "mtime_ns": self.mtime_ns,
            "size_bytes": self.size_bytes,
            "content_hash": self.content_hash,
            "is_deleted": 1 if self.is_deleted else 0,
            "last_seen_at": self.last_seen_at,
            "updated_at": self.updated_at,
            "enrich_state": self.enrich_state,
        }
@dataclass(frozen=True)
class FileEnrichJobDTO:
    job_id: str
    repo_root: str
    relative_path: str
    content_hash: str
    priority: int
    enqueue_source: str
    status: str
    attempt_count: int
    last_error: str | None
    next_retry_at: str
    created_at: str
    updated_at: str
    def to_sql_params(self) -> dict[str, object]:
        return {
            "job_id": self.job_id,
            "repo_root": self.repo_root,
            "relative_path": self.relative_path,
            "content_hash": self.content_hash,
            "priority": self.priority,
            "enqueue_source": self.enqueue_source,
            "status": self.status,
            "attempt_count": self.attempt_count,
            "last_error": self.last_error,
            "next_retry_at": self.next_retry_at,
            "created_at": self.created_at,
            "updated_at": self.updated_at,
        }
@dataclass(frozen=True)
class FileEnrichFailureUpdateDTO:
    job_id: str
    error_message: str
    now_iso: str
    dead_threshold: int
    backoff_base_sec: int
@dataclass(frozen=True)
class EnqueueRequestDTO:
    repo_root: str
    relative_path: str
    content_hash: str
    priority: int
    enqueue_source: str
    now_iso: str
@dataclass(frozen=True)
class CollectedFileBodyDTO:
    repo_root: str
    relative_path: str
    content_hash: str
    content_zlib: bytes
    content_len: int
    normalized_text: str
    created_at: str
    updated_at: str
    def to_sql_params(self) -> dict[str, object]:
        return {
            "repo_root": self.repo_root,
            "relative_path": self.relative_path,
            "content_hash": self.content_hash,
            "content_zlib": self.content_zlib,
            "content_len": self.content_len,
            "normalized_text": self.normalized_text,
            "created_at": self.created_at,
            "updated_at": self.updated_at,
        }
@dataclass(frozen=True)
class ToolReadinessStateDTO:
    repo_root: str
    relative_path: str
    content_hash: str
    list_files_ready: bool
    read_file_ready: bool
    search_symbol_ready: bool
    get_callers_ready: bool
    consistency_ready: bool
    quality_ready: bool
    tool_ready: bool
    last_reason: str
    updated_at: str
    def to_sql_params(self) -> dict[str, object]:
        return {
            "repo_root": self.repo_root,
            "relative_path": self.relative_path,
            "content_hash": self.content_hash,
            "list_files_ready": 1 if self.list_files_ready else 0,
            "read_file_ready": 1 if self.read_file_ready else 0,
            "search_symbol_ready": 1 if self.search_symbol_ready else 0,
            "get_callers_ready": 1 if self.get_callers_ready else 0,
            "consistency_ready": 1 if self.consistency_ready else 0,
            "quality_ready": 1 if self.quality_ready else 0,
            "tool_ready": 1 if self.tool_ready else 0,
            "last_reason": self.last_reason,
            "updated_at": self.updated_at,
        }
@dataclass(frozen=True)
class EnrichStateUpdateDTO:
    repo_root: str
    relative_path: str
    enrich_state: str
    updated_at: str
    def to_sql_params(self) -> dict[str, object]:
        return {
            "repo_root": self.repo_root,
            "relative_path": self.relative_path,
            "enrich_state": self.enrich_state,
            "updated_at": self.updated_at,
        }
@dataclass(frozen=True)
class FileBodyDeleteTargetDTO:
    repo_root: str
    relative_path: str
    content_hash: str
    def to_sql_params(self) -> dict[str, object]:
        return {
            "repo_root": self.repo_root,
            "relative_path": self.relative_path,
            "content_hash": self.content_hash,
        }
@dataclass(frozen=True)
class LspExtractPersistDTO:
    repo_root: str
    relative_path: str
    content_hash: str
    symbols: list[dict[str, object]]
    relations: list[dict[str, object]]
    created_at: str
@dataclass(frozen=True)
class FileListItemDTO:
    repo: str
    relative_path: str
    size_bytes: int
    mtime_ns: int
    content_hash: str
    enrich_state: str
@dataclass(frozen=True)
class FileReadResultDTO:
    relative_path: str
    content: str
    start_line: int
    end_line: int
    source: str
    total_lines: int
    is_truncated: bool
    next_offset: int | None
@dataclass(frozen=True)
class CollectionScanResultDTO:
    scanned_count: int
    indexed_count: int
    deleted_count: int
@dataclass(frozen=True)
class CollectionPolicyDTO:
    include_ext: tuple[str, ...]
    exclude_globs: tuple[str, ...]
    max_file_size_bytes: int
    scan_interval_sec: int
    max_enrich_batch: int
    retry_max_attempts: int
    retry_backoff_base_sec: int
    queue_poll_interval_ms: int
@dataclass(frozen=True)
class PipelineMetricsDTO:
    queue_depth: int
    running_jobs: int
    failed_jobs: int
    dead_jobs: int
    done_jobs: int
    avg_enrich_latency_ms: float
    indexing_mode: str = "steady"
    l2_coverage_bps: int = 0
    l3_coverage_bps: int = 0
    l3_backlog_count: int = 0
    progress_percent_l2: float = 0.0
    progress_percent_l3: float = 0.0
    eta_l2_sec: int = -1
    eta_l3_sec: int = -1
    eta_confidence_bps: int = 0
    eta_window_sec: int = 0
    throughput_ema: float = 0.0
    remaining_jobs_l2: int = 0
    remaining_jobs_l3: int = 0
    worker_state: str = "running"
    last_error_code: str | None = None
    last_error_message: str | None = None
    last_error_at: str | None = None
    def to_dict(self) -> dict[str, object]:
        return {
            "queue_depth": self.queue_depth,
            "running_jobs": self.running_jobs,
            "failed_jobs": self.failed_jobs,
            "dead_jobs": self.dead_jobs,
            "done_jobs": self.done_jobs,
            "avg_enrich_latency_ms": self.avg_enrich_latency_ms,
            "indexing_mode": self.indexing_mode,
            "l2_coverage_bps": self.l2_coverage_bps,
            "l3_coverage_bps": self.l3_coverage_bps,
            "l3_backlog_count": self.l3_backlog_count,
            "progress_percent_l2": self.progress_percent_l2,
            "progress_percent_l3": self.progress_percent_l3,
            "eta_l2_sec": self.eta_l2_sec,
            "eta_l3_sec": self.eta_l3_sec,
            "eta_confidence_bps": self.eta_confidence_bps,
            "eta_window_sec": self.eta_window_sec,
            "throughput_ema": self.throughput_ema,
            "remaining_jobs_l2": self.remaining_jobs_l2,
            "remaining_jobs_l3": self.remaining_jobs_l3,
            "worker_state": self.worker_state,
            "last_error_code": self.last_error_code,
            "last_error_message": self.last_error_message,
            "last_error_at": self.last_error_at,
        }
@dataclass(frozen=True)
class PipelineErrorEventDTO:
    event_id: str
    occurred_at: str
    component: str
    phase: str
    severity: str
    scope_type: str
    repo_root: str | None
    relative_path: str | None
    job_id: str | None
    attempt_count: int
    error_code: str
    error_message: str
    error_type: str
    stacktrace_text: str
    context_json: str
    worker_name: str
    run_mode: str
    resolved: bool
    resolved_at: str | None
    def to_dict(self) -> dict[str, object]:
        return {
            "event_id": self.event_id,
            "occurred_at": self.occurred_at,
            "component": self.component,
            "phase": self.phase,
            "severity": self.severity,
            "scope_type": self.scope_type,
            "repo_root": self.repo_root,
            "relative_path": self.relative_path,
            "job_id": self.job_id,
            "attempt_count": self.attempt_count,
            "error_code": self.error_code,
            "error_message": self.error_message,
            "error_type": self.error_type,
            "stacktrace_text": self.stacktrace_text,
            "context_json": self.context_json,
            "worker_name": self.worker_name,
            "run_mode": self.run_mode,
            "resolved": self.resolved,
            "resolved_at": self.resolved_at,
        }
@dataclass(frozen=True)
class PipelinePolicyDTO:
    deletion_hold: bool
    l3_p95_threshold_ms: int
    dead_ratio_threshold_bps: int
    enrich_worker_count: int
    updated_at: str
    bootstrap_mode_enabled: bool = False
    bootstrap_l3_worker_count: int = 1
    bootstrap_l3_queue_max: int = 1000
    bootstrap_exit_min_l2_coverage_bps: int = 9500
    bootstrap_exit_max_sec: int = 1800
    def to_dict(self) -> dict[str, object]:
        return {
            "deletion_hold": self.deletion_hold,
            "l3_p95_threshold_ms": self.l3_p95_threshold_ms,
            "dead_ratio_threshold_bps": self.dead_ratio_threshold_bps,
            "enrich_worker_count": self.enrich_worker_count,
            "bootstrap_mode_enabled": self.bootstrap_mode_enabled,
            "bootstrap_l3_worker_count": self.bootstrap_l3_worker_count,
            "bootstrap_l3_queue_max": self.bootstrap_l3_queue_max,
            "bootstrap_exit_min_l2_coverage_bps": self.bootstrap_exit_min_l2_coverage_bps,
            "bootstrap_exit_max_sec": self.bootstrap_exit_max_sec,
            "updated_at": self.updated_at,
        }
@dataclass(frozen=True)
class DeadJobItemDTO:
    job_id: str
    repo_root: str
    relative_path: str
    attempt_count: int
    last_error: str | None
    updated_at: str
    def to_dict(self) -> dict[str, object]:
        return {
            "job_id": self.job_id,
            "repo_root": self.repo_root,
            "relative_path": self.relative_path,
            "attempt_count": self.attempt_count,
            "last_error": self.last_error,
            "updated_at": self.updated_at,
        }
@dataclass(frozen=True)
class DeadJobActionResultDTO:
    requeued_count: int = 0
    purged_count: int = 0
    def to_dict(self) -> dict[str, object]:
        return {
            "requeued_count": self.requeued_count,
            "purged_count": self.purged_count,
        }
@dataclass(frozen=True)
class PipelineAlertSnapshotDTO:
    state: str
    window_seconds: int
    event_count: int
    dead_count: int
    dead_ratio_bps: int
    l3_p95_ms: int
    threshold_dead_ratio_bps: int
    threshold_l3_p95_ms: int
    policy: PipelinePolicyDTO
    def to_dict(self) -> dict[str, object]:
        return {
            "state": self.state,
            "window_seconds": self.window_seconds,
            "event_count": self.event_count,
            "dead_count": self.dead_count,
            "dead_ratio_bps": self.dead_ratio_bps,
            "l3_p95_ms": self.l3_p95_ms,
            "threshold_dead_ratio_bps": self.threshold_dead_ratio_bps,
            "threshold_l3_p95_ms": self.threshold_l3_p95_ms,
            "policy": self.policy.to_dict(),
        }
@dataclass(frozen=True)
class PipelineAutoControlStateDTO:
    auto_hold_enabled: bool
    auto_hold_active: bool
    last_action: str
    updated_at: str
    def to_dict(self) -> dict[str, object]:
        return {
            "auto_hold_enabled": self.auto_hold_enabled,
            "auto_hold_active": self.auto_hold_active,
            "last_action": self.last_action,
            "updated_at": self.updated_at,
        }
@dataclass(frozen=True)
class SymbolSearchItemDTO:
    repo: str
    relative_path: str
    name: str
    kind: str
    line: int
    end_line: int
    content_hash: str
    symbol_key: str | None = None
    parent_symbol_key: str | None = None
    depth: int = 0
    container_name: str | None = None
    def to_dict(self) -> dict[str, object]:
        return {
            "repo": self.repo,
            "relative_path": self.relative_path,
            "name": self.name,
            "kind": self.kind,
            "line": self.line,
            "end_line": self.end_line,
            "content_hash": self.content_hash,
            "symbol_key": self.symbol_key,
            "parent_symbol_key": self.parent_symbol_key,
            "depth": self.depth,
            "container_name": self.container_name,
        }
@dataclass(frozen=True)
class CallerEdgeDTO:
    repo: str
    relative_path: str
    from_symbol: str
    to_symbol: str
    line: int
    content_hash: str
    def to_dict(self) -> dict[str, object]:
        return {
            "repo": self.repo,
            "relative_path": self.relative_path,
            "from_symbol": self.from_symbol,
            "to_symbol": self.to_symbol,
            "line": self.line,
            "content_hash": self.content_hash,
        }
@dataclass(frozen=True)
class L3ReferenceDataDTO:
    symbols: list[dict[str, object]]
    relations: list[dict[str, object]]
    error_message: str | None
    def has_error(self) -> bool:
        return self.error_message is not None and self.error_message.strip() != ""
@dataclass(frozen=True)
class L3DiffResultDTO:
    symbol_tp: int
    symbol_fp: int
    symbol_fn: int
    caller_tp: int
    caller_fp: int
    caller_fn: int
    error_message: str | None
    def to_dict(self) -> dict[str, object]:
        return {
            "symbol_tp": self.symbol_tp,
            "symbol_fp": self.symbol_fp,
            "symbol_fn": self.symbol_fn,
            "caller_tp": self.caller_tp,
            "caller_fp": self.caller_fp,
            "caller_fn": self.caller_fn,
            "error_message": self.error_message,
        }
@dataclass(frozen=True)
class SnippetSaveDTO:
    repo_root: str
    source_path: str
    start_line: int
    end_line: int
    tag: str
    note: str | None
    commit_hash: str | None
    content_text: str
    created_at: str
    def to_sql_params(self) -> dict[str, object]:
        return {
            "repo_root": self.repo_root,
            "source_path": self.source_path,
            "start_line": self.start_line,
            "end_line": self.end_line,
            "tag": self.tag,
            "note": self.note,
            "commit_hash": self.commit_hash,
            "content_text": self.content_text,
            "created_at": self.created_at,
        }
@dataclass(frozen=True)
class SnippetRecordDTO:
    snippet_id: int
    repo_root: str
    source_path: str
    start_line: int
    end_line: int
    tag: str
    note: str | None
    commit_hash: str | None
    content_text: str
    created_at: str
    def to_dict(self) -> dict[str, object]:
        return {
            "snippet_id": self.snippet_id,
            "repo": self.repo_root,
            "path": self.source_path,
            "start_line": self.start_line,
            "end_line": self.end_line,
            "tag": self.tag,
            "note": self.note,
            "commit": self.commit_hash,
            "content": self.content_text,
            "created_at": self.created_at,
        }
@dataclass(frozen=True)
class KnowledgeEntryDTO:
    kind: str
    repo_root: str
    topic: str
    content_text: str
    tags: tuple[str, ...]
    related_files: tuple[str, ...]
    created_at: str
    def to_sql_params(self) -> dict[str, object]:
        return {
            "kind": self.kind,
            "repo_root": self.repo_root,
            "topic": self.topic,
            "content_text": self.content_text,
            "tags_json": json.dumps(list(self.tags), ensure_ascii=False),
            "related_files_json": json.dumps(list(self.related_files), ensure_ascii=False),
            "created_at": self.created_at,
        }
@dataclass(frozen=True)
class KnowledgeRecordDTO:
    entry_id: int
    kind: str
    repo_root: str
    topic: str
    content_text: str
    tags: tuple[str, ...]
    related_files: tuple[str, ...]
    created_at: str
    def to_dict(self) -> dict[str, object]:
        return {
            "entry_id": self.entry_id,
            "kind": self.kind,
            "repo": self.repo_root,
            "topic": self.topic,
            "content": self.content_text,
            "tags": list(self.tags),
            "related_files": list(self.related_files),
            "created_at": self.created_at,
        }
def now_iso8601_utc() -> str:
    return datetime.now(timezone.utc).isoformat()



"""stabilization 경고 sink를 제공한다."""

from __future__ import annotations

import threading
from collections import deque


class WarningSink:
    """최근 경고 메시지를 보관하는 경량 sink다."""

    def __init__(self, max_items: int = 512) -> None:
        """보관 상한을 설정한다."""
        self._max_items = max(1, int(max_items))
        self._items: deque[str] = deque()
        self._lock = threading.RLock()

    def add(self, message: str) -> None:
        """경고 메시지를 추가한다."""
        with self._lock:
            self._items.append(str(message))
            while len(self._items) > self._max_items:
                self._items.popleft()

    def list_recent(self, limit: int = 50) -> list[str]:
        """최근 경고 목록을 반환한다."""
        with self._lock:
            sliced = list(self._items)[-max(1, int(limit)) :]
        return sliced


warning_sink = WarningSink()


def warn(message: str) -> None:
    """전역 warning sink에 메시지를 기록한다."""
    warning_sink.add(message)




"""
Provides Kotlin specific instantiation of the LanguageServer class. Contains various configurations and settings specific to Kotlin.

You can configure the following options in ls_specific_settings (in serena_config.yml):

    ls_specific_settings:
      kotlin:
        jvm_options: '-Xmx4G'  # JVM options for Kotlin Language Server (default: -Xmx4G)

Example configuration for large projects:

    ls_specific_settings:
      kotlin:
        jvm_options: '-Xmx8G -XX:+UseG1GC'
"""

import dataclasses
import logging
import os
import pathlib
import stat
from typing import cast

from solidlsp.ls import SolidLanguageServer
from solidlsp.ls_config import Language, LanguageServerConfig
from solidlsp.ls_utils import FileUtils, PlatformUtils
from solidlsp.lsp_protocol_handler.lsp_types import InitializeParams
from solidlsp.lsp_protocol_handler.server import ProcessLaunchInfo
from solidlsp.settings import SolidLSPSettings

log = logging.getLogger(__name__)

# Default JVM options for Kotlin Language Server
# -Xmx4G: Limit max heap to 4GB to prevent OOM on large projects
DEFAULT_KOTLIN_JVM_OPTIONS = "-Xmx4G"


@dataclasses.dataclass
class KotlinRuntimeDependencyPaths:
    """
    Stores the paths to the runtime dependencies of Kotlin Language Server
    """

    java_path: str
    java_home_path: str
    kotlin_executable_path: str


class KotlinLanguageServer(SolidLanguageServer):
    """
    Provides Kotlin specific instantiation of the LanguageServer class. Contains various configurations and settings specific to Kotlin.
    """

    def __init__(self, config: LanguageServerConfig, repository_root_path: str, solidlsp_settings: SolidLSPSettings):
        """
        Creates a Kotlin Language Server instance. This class is not meant to be instantiated directly. Use LanguageServer.create() instead.
        """
        runtime_dependency_paths = self._setup_runtime_dependencies(config, solidlsp_settings)
        self.runtime_dependency_paths = runtime_dependency_paths

        # Create command to execute the Kotlin Language Server script
        cmd = [self.runtime_dependency_paths.kotlin_executable_path, "--stdio"]

        # Get JVM options from settings or use default
        jvm_options = DEFAULT_KOTLIN_JVM_OPTIONS
        if solidlsp_settings.ls_specific_settings:
            kotlin_settings = solidlsp_settings.get_ls_specific_settings(Language.KOTLIN)
            custom_jvm_options = kotlin_settings.get("jvm_options", "")
            if custom_jvm_options:
                jvm_options = custom_jvm_options
                log.info(f"Using custom JVM options for Kotlin Language Server: {jvm_options}")

        # Set environment variables including JAVA_HOME and JVM options
        # JAVA_TOOL_OPTIONS is automatically picked up by any Java process
        proc_env = {
            "JAVA_HOME": self.runtime_dependency_paths.java_home_path,
            "JAVA_TOOL_OPTIONS": jvm_options,
        }

        super().__init__(
            config, repository_root_path, ProcessLaunchInfo(cmd=cmd, env=proc_env, cwd=repository_root_path), "kotlin", solidlsp_settings
        )

    def _requires_open_file_for_document_symbols(self) -> bool:
        return False

    @classmethod
    def _setup_runtime_dependencies(cls, config: LanguageServerConfig, solidlsp_settings: SolidLSPSettings) -> KotlinRuntimeDependencyPaths:
        """
        Setup runtime dependencies for Kotlin Language Server and return the paths.
        """
        platform_id = PlatformUtils.get_platform_id()

        # Verify platform support
        assert (
            platform_id.value.startswith("win-") or platform_id.value.startswith("linux-") or platform_id.value.startswith("osx-")
        ), "Only Windows, Linux and macOS platforms are supported for Kotlin in multilspy at the moment"

        # Runtime dependency information
        runtime_dependencies = {
            "runtimeDependency": {
                "id": "KotlinLsp",
                "description": "Kotlin Language Server",
                "url": "https://download-cdn.jetbrains.com/kotlin-lsp/0.253.10629/kotlin-0.253.10629.zip",
                "archiveType": "zip",
            },
            "java": {
                "win-x64": {
                    "url": "https://github.com/redhat-developer/vscode-java/releases/download/v1.42.0/java-win32-x64-1.42.0-561.vsix",
                    "archiveType": "zip",
                    "java_home_path": "extension/jre/21.0.7-win32-x86_64",
                    "java_path": "extension/jre/21.0.7-win32-x86_64/bin/java.exe",
                },
                "linux-x64": {
                    "url": "https://github.com/redhat-developer/vscode-java/releases/download/v1.42.0/java-linux-x64-1.42.0-561.vsix",
                    "archiveType": "zip",
                    "java_home_path": "extension/jre/21.0.7-linux-x86_64",
                    "java_path": "extension/jre/21.0.7-linux-x86_64/bin/java",
                },
                "linux-arm64": {
                    "url": "https://github.com/redhat-developer/vscode-java/releases/download/v1.42.0/java-linux-arm64-1.42.0-561.vsix",
                    "archiveType": "zip",
                    "java_home_path": "extension/jre/21.0.7-linux-aarch64",
                    "java_path": "extension/jre/21.0.7-linux-aarch64/bin/java",
                },
                "osx-x64": {
                    "url": "https://github.com/redhat-developer/vscode-java/releases/download/v1.42.0/java-darwin-x64-1.42.0-561.vsix",
                    "archiveType": "zip",
                    "java_home_path": "extension/jre/21.0.7-macosx-x86_64",
                    "java_path": "extension/jre/21.0.7-macosx-x86_64/bin/java",
                },
                "osx-arm64": {
                    "url": "https://github.com/redhat-developer/vscode-java/releases/download/v1.42.0/java-darwin-arm64-1.42.0-561.vsix",
                    "archiveType": "zip",
                    "java_home_path": "extension/jre/21.0.7-macosx-aarch64",
                    "java_path": "extension/jre/21.0.7-macosx-aarch64/bin/java",
                },
            },
        }

        kotlin_dependency = runtime_dependencies["runtimeDependency"]
        java_dependency = runtime_dependencies["java"][platform_id.value]  # type: ignore

        # Setup paths for dependencies
        static_dir = os.path.join(cls.ls_resources_dir(solidlsp_settings), "kotlin_language_server")
        os.makedirs(static_dir, exist_ok=True)

        # Setup Java paths
        java_dir = os.path.join(static_dir, "java")
        os.makedirs(java_dir, exist_ok=True)

        java_home_path = os.path.join(java_dir, java_dependency["java_home_path"])
        java_path = os.path.join(java_dir, java_dependency["java_path"])

        # Download and extract Java if not exists
        if not os.path.exists(java_path):
            log.info(f"Downloading Java for {platform_id.value}...")
            FileUtils.download_and_extract_archive(java_dependency["url"], java_dir, java_dependency["archiveType"])
            # Make Java executable
            if not platform_id.value.startswith("win-"):
                os.chmod(java_path, 0o755)

        assert os.path.exists(java_path), f"Java executable not found at {java_path}"

        # Setup Kotlin Language Server paths
        kotlin_ls_dir = static_dir

        # Get platform-specific executable script path
        if platform_id.value.startswith("win-"):
            kotlin_script = os.path.join(kotlin_ls_dir, "kotlin-lsp.cmd")
        else:
            kotlin_script = os.path.join(kotlin_ls_dir, "kotlin-lsp.sh")

        # Download and extract Kotlin Language Server if script doesn't exist
        if not os.path.exists(kotlin_script):
            log.info("Downloading Kotlin Language Server...")
            FileUtils.download_and_extract_archive(kotlin_dependency["url"], static_dir, kotlin_dependency["archiveType"])  # type: ignore

            # Make script executable on Unix platforms
            if os.path.exists(kotlin_script) and not platform_id.value.startswith("win-"):
                os.chmod(
                    kotlin_script, stat.S_IRUSR | stat.S_IWUSR | stat.S_IXUSR | stat.S_IRGRP | stat.S_IXGRP | stat.S_IROTH | stat.S_IXOTH
                )

        # Use script file
        if os.path.exists(kotlin_script):
            kotlin_executable_path = kotlin_script
            log.info(f"Using Kotlin Language Server script at {kotlin_script}")
        else:
            raise FileNotFoundError(f"Kotlin Language Server script not found at {kotlin_script}")

        return KotlinRuntimeDependencyPaths(
            java_path=java_path, java_home_path=java_home_path, kotlin_executable_path=kotlin_executable_path
        )

    @staticmethod
    def _get_initialize_params(repository_absolute_path: str) -> InitializeParams:
        """
        Returns the initialize params for the Kotlin Language Server.
        """
        if not os.path.isabs(repository_absolute_path):
            repository_absolute_path = os.path.abspath(repository_absolute_path)

        root_uri = pathlib.Path(repository_absolute_path).as_uri()
        initialize_params = {
            "clientInfo": {"name": "Multilspy Kotlin Client", "version": "1.0.0"},
            "locale": "en",
            "rootPath": repository_absolute_path,
            "rootUri": root_uri,
            "capabilities": {
                "workspace": {
                    "applyEdit": True,
                    "workspaceEdit": {
                        "documentChanges": True,
                        "resourceOperations": ["create", "rename", "delete"],
                        "failureHandling": "textOnlyTransactional",
                        "normalizesLineEndings": True,
                        "changeAnnotationSupport": {"groupsOnLabel": True},
                    },
                    "didChangeConfiguration": {"dynamicRegistration": True},
                    "didChangeWatchedFiles": {"dynamicRegistration": True, "relativePatternSupport": True},
                    "symbol": {
                        "dynamicRegistration": True,
                        "symbolKind": {"valueSet": list(range(1, 27))},
                        "tagSupport": {"valueSet": [1]},
                        "resolveSupport": {"properties": ["location.range"]},
                    },
                    "codeLens": {"refreshSupport": True},
                    "executeCommand": {"dynamicRegistration": True},
                    "configuration": True,
                    "workspaceFolders": True,
                    "semanticTokens": {"refreshSupport": True},
                    "fileOperations": {
                        "dynamicRegistration": True,
                        "didCreate": True,
                        "didRename": True,
                        "didDelete": True,
                        "willCreate": True,
                        "willRename": True,
                        "willDelete": True,
                    },
                    "inlineValue": {"refreshSupport": True},
                    "inlayHint": {"refreshSupport": True},
                    "diagnostics": {"refreshSupport": True},
                },
                "textDocument": {
                    "publishDiagnostics": {
                        "relatedInformation": True,
                        "versionSupport": False,
                        "tagSupport": {"valueSet": [1, 2]},
                        "codeDescriptionSupport": True,
                        "dataSupport": True,
                    },
                    "synchronization": {"dynamicRegistration": True, "willSave": True, "willSaveWaitUntil": True, "didSave": True},
                    "completion": {
                        "dynamicRegistration": True,
                        "contextSupport": True,
                        "completionItem": {
                            "snippetSupport": False,
                            "commitCharactersSupport": True,
                            "documentationFormat": ["markdown", "plaintext"],
                            "deprecatedSupport": True,
                            "preselectSupport": True,
                            "tagSupport": {"valueSet": [1]},
                            "insertReplaceSupport": False,
                            "resolveSupport": {"properties": ["documentation", "detail", "additionalTextEdits"]},
                            "insertTextModeSupport": {"valueSet": [1, 2]},
                            "labelDetailsSupport": True,
                        },
                        "insertTextMode": 2,
                        "completionItemKind": {
                            "valueSet": [1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25]
                        },
                        "completionList": {"itemDefaults": ["commitCharacters", "editRange", "insertTextFormat", "insertTextMode"]},
                    },
                    "hover": {"dynamicRegistration": True, "contentFormat": ["markdown", "plaintext"]},
                    "signatureHelp": {
                        "dynamicRegistration": True,
                        "signatureInformation": {
                            "documentationFormat": ["markdown", "plaintext"],
                            "parameterInformation": {"labelOffsetSupport": True},
                            "activeParameterSupport": True,
                        },
                        "contextSupport": True,
                    },
                    "definition": {"dynamicRegistration": True, "linkSupport": True},
                    "references": {"dynamicRegistration": True},
                    "documentHighlight": {"dynamicRegistration": True},
                    "documentSymbol": {
                        "dynamicRegistration": True,
                        "symbolKind": {"valueSet": list(range(1, 27))},
                        "hierarchicalDocumentSymbolSupport": True,
                        "tagSupport": {"valueSet": [1]},
                        "labelSupport": True,
                    },
                    "codeAction": {
                        "dynamicRegistration": True,
                        "isPreferredSupport": True,
                        "disabledSupport": True,
                        "dataSupport": True,
                        "resolveSupport": {"properties": ["edit"]},
                        "codeActionLiteralSupport": {
                            "codeActionKind": {
                                "valueSet": [
                                    "",
                                    "quickfix",
                                    "refactor",
                                    "refactor.extract",
                                    "refactor.inline",
                                    "refactor.rewrite",
                                    "source",
                                    "source.organizeImports",
                                ]
                            }
                        },
                        "honorsChangeAnnotations": False,
                    },
                    "codeLens": {"dynamicRegistration": True},
                    "formatting": {"dynamicRegistration": True},
                    "rangeFormatting": {"dynamicRegistration": True},
                    "onTypeFormatting": {"dynamicRegistration": True},
                    "rename": {
                        "dynamicRegistration": True,
                        "prepareSupport": True,
                        "prepareSupportDefaultBehavior": 1,
                        "honorsChangeAnnotations": True,
                    },
                    "documentLink": {"dynamicRegistration": True, "tooltipSupport": True},
                    "typeDefinition": {"dynamicRegistration": True, "linkSupport": True},
                    "implementation": {"dynamicRegistration": True, "linkSupport": True},
                    "colorProvider": {"dynamicRegistration": True},
                    "foldingRange": {
                        "dynamicRegistration": True,
                        "rangeLimit": 5000,
                        "lineFoldingOnly": True,
                        "foldingRangeKind": {"valueSet": ["comment", "imports", "region"]},
                        "foldingRange": {"collapsedText": False},
                    },
                    "declaration": {"dynamicRegistration": True, "linkSupport": True},
                    "selectionRange": {"dynamicRegistration": True},
                    "callHierarchy": {"dynamicRegistration": True},
                    "semanticTokens": {
                        "dynamicRegistration": True,
                        "tokenTypes": [
                            "namespace",
                            "type",
                            "class",
                            "enum",
                            "interface",
                            "struct",
                            "typeParameter",
                            "parameter",
                            "variable",
                            "property",
                            "enumMember",
                            "event",
                            "function",
                            "method",
                            "macro",
                            "keyword",
                            "modifier",
                            "comment",
                            "string",
                            "number",
                            "regexp",
                            "operator",
                            "decorator",
                        ],
                        "tokenModifiers": [
                            "declaration",
                            "definition",
                            "readonly",
                            "static",
                            "deprecated",
                            "abstract",
                            "async",
                            "modification",
                            "documentation",
                            "defaultLibrary",
                        ],
                        "formats": ["relative"],
                        "requests": {"range": True, "full": {"delta": True}},
                        "multilineTokenSupport": False,
                        "overlappingTokenSupport": False,
                        "serverCancelSupport": True,
                        "augmentsSyntaxTokens": True,
                    },
                    "linkedEditingRange": {"dynamicRegistration": True},
                    "typeHierarchy": {"dynamicRegistration": True},
                    "inlineValue": {"dynamicRegistration": True},
                    "inlayHint": {
                        "dynamicRegistration": True,
                        "resolveSupport": {"properties": ["tooltip", "textEdits", "label.tooltip", "label.location", "label.command"]},
                    },
                    "diagnostic": {"dynamicRegistration": True, "relatedDocumentSupport": False},
                },
                "window": {
                    "showMessage": {"messageActionItem": {"additionalPropertiesSupport": True}},
                    "showDocument": {"support": True},
                    "workDoneProgress": True,
                },
                "general": {
                    "staleRequestSupport": {
                        "cancel": True,
                        "retryOnContentModified": [
                            "textDocument/semanticTokens/full",
                            "textDocument/semanticTokens/range",
                            "textDocument/semanticTokens/full/delta",
                        ],
                    },
                    "regularExpressions": {"engine": "ECMAScript", "version": "ES2020"},
                    "markdown": {"parser": "marked", "version": "1.1.0"},
                    "positionEncodings": ["utf-16"],
                },
                "notebookDocument": {"synchronization": {"dynamicRegistration": True, "executionSummarySupport": True}},
            },
            "initializationOptions": {
                "workspaceFolders": [root_uri],
                "storagePath": None,
                "codegen": {"enabled": False},
                "compiler": {"jvm": {"target": "default"}},
                "completion": {"snippets": {"enabled": True}},
                "diagnostics": {"enabled": True, "level": 4, "debounceTime": 250},
                "scripts": {"enabled": True, "buildScriptsEnabled": True},
                "indexing": {"enabled": True},
                "externalSources": {"useKlsScheme": False, "autoConvertToKotlin": False},
                "inlayHints": {"typeHints": False, "parameterHints": False, "chainedHints": False},
                "formatting": {
                    "formatter": "ktfmt",
                    "ktfmt": {
                        "style": "google",
                        "indent": 4,
                        "maxWidth": 100,
                        "continuationIndent": 8,
                        "removeUnusedImports": True,
                    },
                },
            },
            "trace": "verbose",
            "processId": os.getpid(),
            "workspaceFolders": [
                {
                    "uri": root_uri,
                    "name": os.path.basename(repository_absolute_path),
                }
            ],
        }
        return cast(InitializeParams, initialize_params)

    def _start_server(self) -> None:
        """
        Starts the Kotlin Language Server
        """

        def execute_client_command_handler(params: dict) -> list:
            return []

        def do_nothing(params: dict) -> None:
            return

        def window_log_message(msg: dict) -> None:
            log.info(f"LSP: window/logMessage: {msg}")

        self.server.on_request("client/registerCapability", do_nothing)
        self.server.on_notification("language/status", do_nothing)
        self.server.on_notification("window/logMessage", window_log_message)
        self.server.on_request("workspace/executeClientCommand", execute_client_command_handler)
        self.server.on_notification("$/progress", do_nothing)
        self.server.on_notification("textDocument/publishDiagnostics", do_nothing)
        self.server.on_notification("language/actionableNotification", do_nothing)

        log.info("Starting Kotlin server process")
        self.server.start()
        initialize_params = self._get_initialize_params(self.repository_root_path)

        log.info("Sending initialize request from LSP client to LSP server and awaiting response")
        init_response = self.server.send.initialize(initialize_params)

        capabilities = init_response["capabilities"]
        assert "textDocumentSync" in capabilities, "Server must support textDocumentSync"
        assert "hoverProvider" in capabilities, "Server must support hover"
        assert "completionProvider" in capabilities, "Server must support code completion"
        assert "signatureHelpProvider" in capabilities, "Server must support signature help"
        assert "definitionProvider" in capabilities, "Server must support go to definition"
        assert "referencesProvider" in capabilities, "Server must support find references"
        assert "documentSymbolProvider" in capabilities, "Server must support document symbols"
        assert "workspaceSymbolProvider" in capabilities, "Server must support workspace symbols"
        assert "semanticTokensProvider" in capabilities, "Server must support semantic tokens"

        self.server.notify.initialized({})
